from unittest.mock import call, MagicMock, patch

import fedora_messaging.api
import pytest

from toddlers.exceptions import PagureError
import toddlers.plugins.clean_retired_packages


class TestCleanRetiredPackagesToddler:
    toddler_cls = toddlers.plugins.clean_retired_packages.CleanRetiredPackages

    def test_accepts_topic_invalid(self, toddler):
        assert toddler.accepts_topic("foo.bar") is False

    @pytest.mark.parametrize(
        "topic",
        [
            "org.fedoraproject.*.toddlers.trigger.clean_retired_packages",
            "org.fedoraproject.prod.toddlers.trigger.clean_retired_packages",
            "org.fedoraproject.stg.toddlers.trigger.clean_retired_packages",
        ],
    )
    def test_accepts_topic_valid(self, topic, toddler):
        assert toddler.accepts_topic(topic)

    @patch("toddlers.utils.notify.send_email")
    @patch("toddlers.utils.pagure.set_pagure")
    @patch("toddlers.utils.bodhi.set_bodhi")
    def test_process(
        self, mock_set_bodhi, mock_set_pagure, send_email, toddler, caplog
    ):
        mock_dist_git = MagicMock()
        mock_dist_git.get_retired_packages.side_effect = (
            ["package01", "package02"],
            ["package01", "package03"],
        )
        mock_dist_git.get_branches.return_value = ["f39", "f40"]
        mock_dist_git.get_project.return_value = {
            "user": {"name": "Gavriel Loken"},
            "access_users": {"admin": ["Fulgrim", "orphan"]},
            "access_groups": {"admin": ["Adeptus Astartes"]},
        }
        mock_set_pagure.return_value = mock_dist_git
        mock_bodhi = MagicMock()
        mock_bodhi.get_active_branches.return_value = ["f39", "f40"]
        mock_set_bodhi.return_value = mock_bodhi

        namespace = "rpms"

        msg = fedora_messaging.api.Message()
        msg.id = 123
        msg.topic = "org.fedoraproject.prod.toddlers.trigger.clean_retired_packages"
        msg.body = {"foo": "bar"}

        config = {
            "dist_git_url": "https://src.fedoraproject.org/",
            "dist_git_token": "foobar",
            "admin_email": "admin@fp.o",
            "mail_server": "mail_server",
        }

        assert toddler.process(config=config, message=msg) is None

        mock_bodhi.get_active_branches.assert_called_once()
        mock_dist_git.get_retired_packages.assert_has_calls(
            calls=[call("f39"), call("f40")]
        )
        mock_dist_git.get_branches.assert_has_calls(
            calls=[
                call(namespace, "package01"),
                call(namespace, "package02"),
                call(namespace, "package03"),
            ]
        )
        mock_dist_git.get_project.assert_called_with(namespace, "package01")
        mock_dist_git.orphan_package.assert_called_with(
            namespace,
            "package01",
            "Other",
            "Package was retired before it was orphaned",
        )
        mock_dist_git.set_bugzilla_overrides.assert_called_with(
            namespace, "package01", None, None
        )
        mock_dist_git.modify_acls.assert_has_calls(
            calls=[
                call(namespace, "package01", "user", "Fulgrim"),
                call(namespace, "package01", "group", "Adeptus Astartes"),
            ]
        )
        mock_dist_git.update_watcher_status.assert_called_with(
            namespace, "package01", -1, "Fulgrim"
        )
        send_email.assert_called_with(
            to_addresses=["admin@fp.o"],
            from_address="admin@fp.o",
            subject="Toddlers cleaned up some retired packages",
            content="Dear Admin,\n\n"
            "The clean_retired_packages toddler just ran and adjusted some retired packages\n"
            "for recording purposes, here is what it did:\n"
            "- Orphaning rpms/package01 from Gavriel Loken\n"
            "- Removing access on rpms/package01 to Fulgrim\n"
            "- Removing access on rpms/package01 to @Adeptus Astartes\n\n"
            "In case someone asks, you can come back to this email but there isn't really\n"
            "anything else to do with it.\n\n"
            "Have a wonderful day and see you (maybe?) at the next run!\n\n",
            mail_server="mail_server",
        )

    @patch("toddlers.utils.notify.send_email")
    @patch("toddlers.utils.pagure.set_pagure")
    @patch("toddlers.utils.bodhi.set_bodhi")
    def test_process_pagure_repo_erorr(
        self, mock_set_bodhi, mock_set_pagure, send_email, toddler, caplog
    ):
        mock_dist_git = MagicMock()
        mock_dist_git.get_retired_packages.side_effect = (["package01"],)
        mock_dist_git.get_branches.side_effect = PagureError("Error")
        mock_set_pagure.return_value = mock_dist_git
        mock_bodhi = MagicMock()
        mock_bodhi.get_active_branches.return_value = ["f40"]
        mock_set_bodhi.return_value = mock_bodhi

        namespace = "rpms"

        msg = fedora_messaging.api.Message()
        msg.id = 123
        msg.topic = "org.fedoraproject.prod.toddlers.trigger.clean_retired_packages"
        msg.body = {"foo": "bar"}

        config = {
            "dist_git_url": "https://src.fedoraproject.org/",
            "dist_git_token": "foobar",
            "admin_email": "admin@fp.o",
            "mail_server": "mail_server",
        }

        assert toddler.process(config=config, message=msg) is None

        mock_bodhi.get_active_branches.assert_called_once()
        mock_dist_git.get_retired_packages.assert_has_calls(calls=[call("f40")])
        mock_dist_git.get_branches.assert_called_with(namespace, "package01")
        mock_dist_git.get_project.assert_not_called()
        mock_dist_git.orphan_package.assert_not_called()
        mock_dist_git.set_bugzilla_overrides.assert_not_called()
        mock_dist_git.modify_acls.assert_not_called()
        mock_dist_git.update_watcher_status.assert_not_called()
        send_email.assert_not_called()
