import logging
import re
from unittest.mock import ANY, patch

import fedora_messaging.testing
import pytest

import toddlers.playtime


class TestPlaytime:
    def test_playtime_no_args(self, capsys):
        with pytest.raises(SystemExit):
            toddlers.playtime.main([])

        out, err = capsys.readouterr()
        assert out == ""
        # Expecting something along these lines, but don't make the test too tight:
        #
        # usage: pytest [-h] [--conf CONF] [--debug] name
        # pytest: error: the following arguments are required: name
        assert err.startswith("usage:")
        assert "error: the following arguments are required:" in err

    def test_playtime_arg(self, capsys):
        with fedora_messaging.testing.mock_sends(
            fedora_messaging.api.Message(
                topic="toddlers.trigger.debug", body={"date": ANY}
            )
        ):
            toddlers.playtime.main(["debug"])
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""

    def test_playtime_invalid_config(self, capsys):
        with fedora_messaging.testing.mock_sends(
            fedora_messaging.api.Message(
                topic="toddlers.trigger.debug", body={"date": ANY}
            )
        ):
            toddlers.playtime.main(["debug", "--conf=foo"])
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""

    @pytest.mark.parametrize(
        "exception, message",
        [
            (Exception("foo!"), "Error sending fedora-messaging message: foo!"),
            (
                fedora_messaging.exceptions.PublishReturned("Ahah!"),
                "Fedora Messaging broker rejected message .*: Ahah!",
            ),
            (
                fedora_messaging.exceptions.ConnectionException("Bouh!"),
                "Error sending message .*: Bouh!",
            ),
        ],
    )
    def test_playtime_exceptions(self, exception, message, caplog):
        caplog.set_level(logging.INFO)
        with patch("fedora_messaging.api.publish", side_effect=exception):
            toddlers.playtime.main(["debug"])
            if "*" in message:
                assert re.match(message, caplog.records[-1].message)
            else:
                assert caplog.records[-1].message == message
