from base64 import b64encode
from unittest.mock import Mock, patch

import toddlers.utils.notify


def assert_message_sent(smtp_mock, msg_string, from_addr, to_addrs):
    smtp_mock.send_message.assert_called_once()
    assert smtp_mock.send_message.call_args.args[1] == from_addr
    assert smtp_mock.send_message.call_args.args[2] == to_addrs
    sent_msg = smtp_mock.send_message.call_args.args[0]
    assert sent_msg.as_string() == msg_string


class TestNotify:
    @patch("toddlers.utils.notify.smtplib")
    def test_notify_packager(self, mock_smtp):
        smtp_server = Mock()
        mock_smtp.SMTP.return_value = smtp_server

        expected_msg = """To: pingou@email
From: admin@server
Subject: Fedora Account System and Bugzilla Mismatch
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: quoted-printable
MIME-Version: 1.0

Hello pingou,

We have identified you[1] as either a Fedora packager or someone who has aske=
d to
be included in the CC list of tickets created for one or more component on
bugzilla. Fedora packagers are granted special permissions on the Fedora bugs=
 in
bugzilla.
However, to enable these functionalities (granting you these permissions or
including you to the CC list of your packages of interest), we need to have y=
our
bugzilla email address stored in the Fedora Account System[2].
At the moment you have:

pingou@email

which bugzilla is telling us is not an account in bugzilla.  If you could
please set up an account in bugzilla with this address or change your email
address (either Red Hat Bugzilla email or Fedora Account email) on your Fedora
Account to match an existing bugzilla account this would let us go forward.

Note #1: this message is being generated by an automated script.  You'll
continue getting this message until the problem is resolved.  Sorry for the
inconvenience.

Thank you,
The Fedora Account System
admin@server


[1] the source of this information is the following JSON file:
    https://src.fedoraproject.org/extras/pagure_bz.json
    We are happy to tell you exactly which packages are linked to your account
    if you wish.
[2] https://accounts.fedoraproject.org/
"""

        toddlers.utils.notify.notify_packager(
            mail_server="server.mail",
            admin_email="admin@server",
            username="pingou",
            email="pingou@email",
        )
        mock_smtp.SMTP.assert_called_with("server.mail")
        assert_message_sent(smtp_server, expected_msg, "admin@server", ["pingou@email"])

    @patch("toddlers.utils.notify.smtplib")
    def test_notify_admins(self, mock_smtp):
        users = []
        for name, email in [
            ("pingou", "pingou@mail"),
            ("nils", "nils@mail"),
            ("trasher", "trasher@mail"),
        ]:
            user = Mock()
            if name == "trasher":
                user.person.status = "Inactive"
            else:
                user.person.status = "Active"
            user.person.username = name
            user.person.human_name = name
            user.email = email
            users.append(user)

        smtp_server = Mock()
        mock_smtp.SMTP.return_value = smtp_server

        expected_msg = """To: info_admin@server
From: admin@server
Subject: Fedora Account System and Bugzilla Mismatch
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: 7bit
MIME-Version: 1.0


The following people are in the packager group but do not have email addresses
that are valid in bugzilla:
  pingou  --  pingou  --  pingou@mail
  nils  --  nils  --  nils@mail

"""

        toddlers.utils.notify.notify_admins(
            mail_server="server.mail",
            admin_email="admin@server",
            recipients=["info_admin@server"],
            users=users,
        )
        mock_smtp.SMTP.assert_called_with("server.mail")
        assert_message_sent(
            smtp_server, expected_msg, "admin@server", ["info_admin@server"]
        )

    @patch("toddlers.utils.notify.smtplib")
    def test_notify_admins_distgit_sync_error(self, mock_smtp):
        smtp_server = Mock()
        mock_smtp.SMTP.return_value = smtp_server

        expected_msg = """To: info_admin@server
From: admin@server
Subject: Errors while syncing bugzilla with the PackageDB
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: quoted-printable
MIME-Version: 1.0

Greetings, our mighty admins,

While updating bugzilla with information from the Package Database some error=
s were encountered.
Don't panic!
Please, be so kind and see if you can do something with them. Here is the lis=
t of the errors:

Oops, something happened
Oh no, this is bad
"""

        errors = ["Oops, something happened", "Oh no, this is bad"]

        toddlers.utils.notify.notify_admins_distgit_sync_error(
            mail_server="server.mail",
            admin_email="admin@server",
            recipients=["info_admin@server"],
            errors=errors,
        )
        mock_smtp.SMTP.assert_called_with("server.mail")
        assert_message_sent(
            smtp_server, expected_msg, "admin@server", ["info_admin@server"]
        )

    @patch("toddlers.utils.notify.smtplib")
    def test_notify_send_email_with_cc_addresses(self, mock_smtp):
        smtp_server = Mock()
        mock_smtp.SMTP.return_value = smtp_server

        content = """Greetings, our mighty admins,

This is a test email!

Ahah!
"""

        expected_msg = f"""To: info_admin@server
From: admin@server
Cc: admin@server2
Subject: Test email
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: 7bit
MIME-Version: 1.0

{content}"""
        toddlers.utils.notify.send_email(
            to_addresses=["info_admin@server"],
            from_address="admin@server",
            subject="Test email",
            content=content,
            mail_server="server.mail",
            cc_address=["admin@server2"],
        )
        mock_smtp.SMTP.assert_called_with("server.mail")
        assert_message_sent(
            smtp_server,
            expected_msg,
            "admin@server",
            ["info_admin@server", "admin@server2"],
        )

    @patch("toddlers.utils.notify.smtplib")
    def test_notify_send_email_utf8(self, mock_smtp):
        smtp_server = Mock()
        mock_smtp.SMTP.return_value = smtp_server

        content = "Some non-ascii: Miro Hron\u010dok.\n"
        expected_msg = f"""To: info_admin@server
From: admin@server
Subject: Test email
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: base64
MIME-Version: 1.0

{b64encode(content.encode('utf8')).decode()}
"""
        toddlers.utils.notify.send_email(
            to_addresses=["info_admin@server"],
            from_address="admin@server",
            subject="Test email",
            content=content,
            mail_server="server.mail",
        )
        mock_smtp.SMTP.assert_called_with("server.mail")
        assert_message_sent(
            smtp_server, expected_msg, "admin@server", ["info_admin@server"]
        )
