"""
Unit tests for `toddlers.utils.pagure`.
"""

import json
import logging
from unittest.mock import call, MagicMock, Mock, patch
import urllib

import pytest

from toddlers.exceptions import PagureError
from toddlers.utils.cache import cache
import toddlers.utils.pagure as pagure


class TestPagureSetPagure:
    """
    Test class for `toddlers.pagure.set_pagure` function.
    """

    def test_set_pagure(self):
        """
        Test initialization of pagure module.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        pagure_obj = pagure.set_pagure(config)

        assert pagure_obj._pagure_url == config.get("pagure_url")
        assert pagure_obj._pagure_api_key == config.get("pagure_api_key")
        assert pagure_obj._requests_session

    def test_set_pagure_no_pagure_url(self):
        """
        Test initialization of pagure module without required config value.
        """
        with pytest.raises(
            ValueError, match=r"No pagure_url found in the configuration file"
        ):
            pagure.set_pagure({})

    def test_set_pagure_no_pagure_api_key(self):
        """
        Test initialization of pagure module without required config value.
        """
        with pytest.raises(
            ValueError, match=r"No pagure_api_key found in the configuration file"
        ):
            config = {"pagure_url": "https://pagure.io"}
            pagure.set_pagure(config)


class TestPagureGetAuthHeader:
    """
    Test class for `toddlers.pagure.Pagure.get_auth_header` function.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)

    def test_get_auth_header(self):
        """
        Assert that correct header is returned.
        """
        exp_header = {
            "Authorization": "token Very secret key",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        header = self.pagure.get_auth_header()

        assert exp_header == header


class TestPagureGetIssue:
    """
    Test class for `toddlers.pagure.Pagure.get_issue` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_issue(self):
        """
        Assert that issue on pagure is retrieved correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200
        data = {"id": 100}
        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        issue_id = 100
        namespace = "test"

        result = self.pagure.get_issue(issue_id, namespace)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/test/issue/100",
            headers=self.pagure.get_auth_header(),
        )

        assert result == data

    def test_get_issue_not_ok(self):
        """
        Assert that failing to retrieve issue on pagure is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.get.return_value = response_mock

        issue_id = 100
        namespace = "test"

        expected_error = "Couldn't retrieve issue '{}' from project '{}'".format(
            issue_id, namespace
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_issue(issue_id, namespace)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/test/issue/100",
            headers=self.pagure.get_auth_header(),
        )


class TestPagureCloseIssue:
    """
    Test class for `toddlers.pagure.Pagure.close_issue` function.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_close_issue(self):
        """
        Assert that issue on pagure is closed correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        issue_id = 100
        namespace = "test"
        message = "message"
        reason = "Closed"

        with patch("toddlers.utils.pagure.Pagure.add_comment_to_issue") as comment_mock:
            comment_mock.return_value = True
            self.pagure.close_issue(issue_id, namespace, message, reason)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/test/issue/100/status",
            data=json.dumps({"status": reason}),
            headers=self.pagure.get_auth_header(),
        )

    def test_close_issue_not_ok(self):
        """
        Assert that failing to close issue on pagure is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.return_value = response_mock

        issue_id = 100
        namespace = "test"
        message = "message"
        reason = "Closed"

        expected_error = "Couldn't close issue 'https://pagure.io/test/issue/100'"

        with pytest.raises(PagureError, match=expected_error):
            with patch(
                "toddlers.utils.pagure.Pagure.add_comment_to_issue"
            ) as comment_mock:
                comment_mock.return_value = True
                self.pagure.close_issue(issue_id, namespace, message, reason)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/test/issue/100/status",
            data=json.dumps({"status": reason}),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureAddCommentToIssue:
    """
    Test class for `toddlers.pagure.Pagure.add_comment_to_issue` function.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_add_comment_to_issue(self):
        """
        Assert that comment on pagure issue is posted correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        issue_id = 100
        namespace = "test"
        message = "message"

        self.pagure.add_comment_to_issue(issue_id, namespace, message)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/test/issue/100/comment",
            data=json.dumps({"comment": message}),
            headers=self.pagure.get_auth_header(),
        )

    def test_add_comment_to_issue_not_ok(self):
        """
        Assert that failing to comment on issue on pagure is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.return_value = response_mock

        issue_id = 100
        namespace = "test"
        message = "message"

        expected_error = "Couldn't comment on issue 'https://pagure.io/test/issue/100'"

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.add_comment_to_issue(issue_id, namespace, message)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/test/issue/100/comment",
            data=json.dumps({"comment": message}),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureUserExists:
    """
    Test class for `toddlers.pagure.Pagure.user_exists` function.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_user_exists(self):
        """
        Assert that everything is handled correctly if user exists.
        """
        username = "Khaine"

        response_mock = Mock()
        response_mock.status_code = 200
        data = {"users": [username]}
        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        result = self.pagure.user_exists(username)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/users?pattern={0}".format(username),
        )
        assert result is True

    def test_user_exists_no_user(self):
        """
        Assert that everything is handled correctly if user doesn't exists.
        """
        username = "Khaine"

        response_mock = Mock()
        response_mock.status_code = 200
        data = {"users": []}
        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        result = self.pagure.user_exists(username)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/users?pattern={0}".format(username),
        )
        assert result is False

    def test_user_exists_not_ok(self):
        """
        Assert that failing to get user is handled correctly.
        """
        username = "Khaine"

        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.return_value = response_mock

        expected_error = "Couldn't get user '{0}'".format(username)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.user_exists(username)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/users?pattern={0}".format(username),
        )


class TestPagureNewProject:
    """
    Test class for `toddlers.pagure.Pagure.new_project` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_new_project(self):
        """
        Assert that new project on pagure is created correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        default_branch = "default_branch"
        description = "description"
        upstream_url = "https://example.com"

        self.pagure.new_project(
            namespace, repo, description, upstream_url, default_branch
        )

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/new",
            data=json.dumps(
                {
                    "namespace": namespace,
                    "name": repo,
                    "default_branch": default_branch,
                    "description": description,
                    "url": upstream_url,
                    "wait": True,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_new_project_initial_commit(self):
        """
        Assert that new project on pagure is created correctly with initial commit.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        default_branch = "default_branch"
        description = "description"
        upstream_url = "https://example.com"

        self.pagure.new_project(
            namespace,
            repo,
            description,
            upstream_url,
            default_branch,
            initial_commit=True,
        )

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/new",
            data=json.dumps(
                {
                    "namespace": namespace,
                    "name": repo,
                    "default_branch": default_branch,
                    "description": description,
                    "url": upstream_url,
                    "wait": True,
                    "create_readme": True,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_new_project_alias(self):
        """
        Assert that new project on pagure is created correctly with alias.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        repo = "repo"
        default_branch = "default_branch"
        description = "description"
        upstream_url = ""
        initial_commit = "sha256"

        self.pagure.new_project(
            namespace,
            repo,
            description,
            upstream_url,
            default_branch,
            initial_commit=initial_commit,
            alias=True,
        )

        self.pagure._requests_session.post.assert_has_calls(
            [
                call(
                    "https://pagure.io/api/0/new",
                    data=json.dumps(
                        {
                            "namespace": namespace,
                            "name": repo,
                            "default_branch": default_branch,
                            "description": description,
                            "url": upstream_url,
                            "wait": True,
                            "create_readme": True,
                        }
                    ),
                    headers=self.pagure.get_auth_header(),
                ),
                call(
                    "https://pagure.io/api/0/{0}/{1}/git/alias/new".format(
                        namespace, repo
                    ),
                    data=json.dumps(
                        {
                            "alias_from": "main",
                            "alias_to": "rawhide",
                        }
                    ),
                    headers=self.pagure.get_auth_header(),
                ),
            ]
        )

    def test_new_project_alias_wrong_namespace(self):
        """
        Assert that new project on pagure is created correctly without alias when
        in wrong namespace.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        default_branch = "default_branch"
        description = "description"
        upstream_url = "https://example.com"
        initial_commit = "sha256"

        self.pagure.new_project(
            namespace,
            repo,
            description,
            upstream_url,
            default_branch,
            initial_commit=initial_commit,
            alias=True,
        )

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/new",
            data=json.dumps(
                {
                    "namespace": namespace,
                    "name": repo,
                    "default_branch": default_branch,
                    "description": description,
                    "url": upstream_url,
                    "wait": True,
                    "create_readme": True,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_new_project_failure(self):
        """
        Assert that failing to create project on pagure is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        default_branch = "default_branch"
        description = "description"
        upstream_url = "https://example.com"

        expected_error = "Couldn't create project '{0}/{1}'".format(namespace, repo)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.new_project(
                namespace, repo, description, upstream_url, default_branch
            )

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/new",
            data=json.dumps(
                {
                    "namespace": namespace,
                    "name": repo,
                    "default_branch": default_branch,
                    "description": description,
                    "url": upstream_url,
                    "wait": True,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_new_project_alias_failure(self):
        """
        Assert that failing to create alias on pagure is handled correctly.
        """
        response_mock_ok = Mock()
        response_mock_ok.status_code = 200

        response_mock_fail = Mock()
        response_mock_fail.status_code = 500

        self.pagure._requests_session.post.side_effect = [
            response_mock_ok,
            response_mock_fail,
        ]

        namespace = "rpms"
        repo = "repo"
        default_branch = "default_branch"
        description = "description"
        upstream_url = "https://example.com"
        initial_commit = "sha256"

        expected_error = "Couldn't create alias for project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.new_project(
                namespace,
                repo,
                description,
                upstream_url,
                default_branch,
                initial_commit=initial_commit,
                alias=True,
            )

        self.pagure._requests_session.post.assert_has_calls(
            [
                call(
                    "https://pagure.io/api/0/new",
                    data=json.dumps(
                        {
                            "namespace": namespace,
                            "name": repo,
                            "default_branch": default_branch,
                            "description": description,
                            "url": upstream_url,
                            "wait": True,
                            "create_readme": True,
                        }
                    ),
                    headers=self.pagure.get_auth_header(),
                ),
                call(
                    "https://pagure.io/api/0/{0}/{1}/git/alias/new".format(
                        namespace, repo
                    ),
                    data=json.dumps(
                        {
                            "alias_from": "main",
                            "alias_to": "rawhide",
                        }
                    ),
                    headers=self.pagure.get_auth_header(),
                ),
            ]
        )


class TestPagureNewBranch:
    """
    Test class for `toddlers.pagure.Pagure.new_branch` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_new_branch_from_commit(self):
        """
        Assert that new branch from commit is created correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        branch = "branch"
        from_commit = "commit"

        self.pagure.new_branch(namespace, repo, branch, from_commit=from_commit)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branch".format(namespace, repo),
            data=json.dumps(
                {
                    "branch": branch,
                    "from_commit": from_commit,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_new_branch_from_branch(self):
        """
        Assert that new branch from existing branch is created correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        branch = "branch"
        from_branch = "existing_branch"

        self.pagure.new_branch(namespace, repo, branch, from_branch=from_branch)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branch".format(namespace, repo),
            data=json.dumps(
                {
                    "branch": branch,
                    "from_branch": from_branch,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_new_branch_missing_required_params(self):
        """
        Assert that method fails if required parameters are not provided.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        branch = "branch"

        expected_error = "You must specify either `from_commit` or " "`from_branch`"

        with pytest.raises(RuntimeError, match=expected_error):
            self.pagure.new_branch(namespace, repo, branch)

    def test_new_branch_conflicting_params(self):
        """
        Assert that method fails if conflicting parameters are provided.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        branch = "branch"
        from_branch = "existing_branch"
        from_commit = "commit"

        expected_error = (
            "`from_commit` and `from_branch` were both " "specified. Only use one."
        )

        with pytest.raises(RuntimeError, match=expected_error):
            self.pagure.new_branch(
                namespace,
                repo,
                branch,
                from_branch=from_branch,
                from_commit=from_commit,
            )

    def test_new_branch_failure(self):
        """
        Assert that failing to create branch on pagure is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.side_effect = response_mock

        namespace = "namespace"
        repo = "repo"
        branch = "branch"
        from_branch = "existing_branch"

        expected_error = "Couldn't create branch in project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.new_branch(namespace, repo, branch, from_branch=from_branch)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branch".format(namespace, repo),
            data=json.dumps(
                {
                    "branch": branch,
                    "from_branch": from_branch,
                }
            ),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureSetMonitoringStatus:
    """
    Test class for `toddlers.pagure.Pagure.set_monitoring_status` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_set_monitoring_status(self):
        """
        Assert that setting monitoring status is done correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        monitoring_level = "no_monitoring"

        self.pagure.set_monitoring_status(namespace, repo, monitoring_level)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/_dg/anitya/{0}/{1}".format(namespace, repo),
            data=json.dumps({"anitya_status": monitoring_level}),
            headers=self.pagure.get_auth_header(),
        )

    def test_set_monitoring_status_failure(self):
        """
        Assert that failing to set monitoring level is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.post.side_effect = response_mock

        namespace = "namespace"
        repo = "repo"
        monitoring_level = "no_monitoring"

        expected_error = "Couldn't set monitoring on project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.set_monitoring_status(namespace, repo, monitoring_level)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/_dg/anitya/{0}/{1}".format(namespace, repo),
            data=json.dumps({"anitya_status": monitoring_level}),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureChangeProjectMainAdmin:
    """
    Test class for `toddlers.pagure.Pagure.change_project_main_admin` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_change_project_main_admin(self):
        """
        Assert that changing project main admin is processed correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.patch.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        main_admin = "zlopez"

        self.pagure.change_project_main_admin(namespace, repo, main_admin)

        self.pagure._requests_session.patch.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}".format(namespace, repo),
            data=json.dumps({"main_admin": main_admin}),
            headers=self.pagure.get_auth_header(),
        )

    def test_change_project_main_admin_failure(self):
        """
        Assert that failing to set main admin is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.patch.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        main_admin = "zlopez"

        expected_error = "Couldn't set new admin on project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.change_project_main_admin(namespace, repo, main_admin)

        self.pagure._requests_session.patch.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}".format(namespace, repo),
            data=json.dumps({"main_admin": main_admin}),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetProjectContributors:
    """
    Test class for `toddlers.pagure.Pagure.get_project_contributors` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_project_contributors(self):
        """
        Assert that getting contributors works correctly.
        """
        response_mock = Mock()
        response_mock.ok = True
        response_mock.status_code = 200

        username = "zlopez"
        data = {"username": username}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_project_contributors(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/contributors".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result == data

    def test_get_project_doesnt_exist(self):
        """
        Assert that failing to get contributors when getting contributors
        is handled correctly.
        """
        response_mock = Mock()
        response_mock.ok = False
        response_mock.status_code = 404

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_project_contributors(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/contributors".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result is None

    def test_get_project_contributors_failure(self):
        """
        Assert that failing to get contributors is handling correctly.
        """
        response_mock = Mock()
        response_mock.ok = False
        response_mock.status_code = 500

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        expected_error = "Couldn't get contributors for project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_project_contributors(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/contributors".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetBranches:
    """
    Test class for `toddlers.pagure.Pagure.get_branches` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_branches(self):
        """
        Assert that getting branches works correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        branch = "branch"
        data = {"branches": [branch]}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_branches(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branches".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result == [branch]

    def test_get_branches_no_branches(self):
        """
        Assert that getting branches returns [] if the project doesn't have any branch.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        data = {}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_branches(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branches".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result == []

    def test_get_branches_failure(self):
        """
        Assert that failing to get branches is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.get.side_effect = response_mock

        namespace = "namespace"
        repo = "repo"

        expected_error = "Couldn't get list of branches for project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_branches(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branches".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetDefaultBranch:
    """
    Test class for `toddlers.pagure.Pagure.get_default_branch` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_default_branch(self):
        """
        Assert that getting default branch works correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        branch = "branch"
        data = {"default": branch}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_default_branch(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branches".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result == branch

    def test_get_default_branch_not_set(self):
        """
        Assert that getting default branch returns None if not set.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        data = {}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_default_branch(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branches".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result is None

    def test_get_default_branch_failure(self):
        """
        Assert that failing to get default branch is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.get.side_effect = response_mock

        namespace = "namespace"
        repo = "repo"

        expected_error = "Couldn't get default branch for project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_default_branch(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/branches".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetProject:
    """
    Test class for `toddlers.pagure.Pagure.get_project` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_project(self):
        """
        Assert that getting project_info works correctly.
        """
        response_mock = Mock()
        response_mock.ok = True

        project_name = "zlopez"
        data = {"name": project_name}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_project(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result == data

    def test_get_project_doesnt_exist(self):
        """
        Assert that getting project_info doesn't return anything if the project doesn't exist.
        """
        response_mock = Mock()
        response_mock.ok = False
        response_mock.status_code = 404

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.get_project(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )

        assert result is None

    def test_get_project_failure(self):
        """
        Assert that failing to get project data is handled correctly.
        """
        response_mock = Mock()
        response_mock.ok = False
        response_mock.status_code = 500

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        expected_error = "Couldn't get project '{0}/{1}'".format(namespace, repo)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_project(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}".format(namespace, repo),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureIsProjectOrphaned:
    """
    Test class for `toddlers.pagure.Pagure.is_project_orphaned` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_is_project_orphaned_project_orphaned(self):
        """
        Assert that checking if project is orphaned works correctly. Project is orphaned.
        """
        response_mock = Mock()
        response_mock.ok = True

        keyword = "orphan"
        data = {"epel_assignee": keyword, "fedora_assignee": "some_fas"}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.is_project_orphaned(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/" + namespace + "/" + repo,
            headers=self.pagure.get_auth_header(),
        )

        assert result is True

    def test_is_project_orphaned_project_not_orphaned(self):
        """
        Assert that checking if project is orphaned works correctly. Project is not orphaned.
        """
        response_mock = Mock()
        response_mock.ok = True

        data = {"epel_assignee": "some_fas", "fedora_assignee": "some_fas"}

        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        result = self.pagure.is_project_orphaned(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/" + namespace + "/" + repo,
            headers=self.pagure.get_auth_header(),
        )

        assert result is False

    def test_is_project_orphaned_failure(self):
        """
        Assert that checking if project is orphaned will raise error if the project doesn't exist.
        """
        response_mock = Mock()
        response_mock.ok = False
        response_mock.status_code = 404

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "namespace"
        repo = "repo"

        expected_error = "Couldn't get project '{0}/{1}' maintainers.".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.is_project_orphaned(namespace, repo)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/" + namespace + "/" + repo,
            headers=self.pagure.get_auth_header(),
        )


class TestPagureAssignMaintainerToProject:
    """
    Test class for `toddlers.pagure.Pagure.assign_maintainer_to_project` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_assign_maintainer_to_project(self):
        """
        Assert that assigning maintainer to project is processed correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.patch.return_value = response_mock

        namespace = "namespace"
        repo = "repo"
        maintainer_fas = "amedvede"
        payload = {
            "EPEL Maintainer name": maintainer_fas,
            "Maintainer name": maintainer_fas,
        }

        self.pagure.assign_maintainer_to_project(namespace, repo, maintainer_fas)

        self.pagure._requests_session.patch.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/" + namespace + "/" + repo,
            data=json.dumps(payload),
            headers=self.pagure.get_auth_header(),
        )

    def test_assign_maintainer_to_project_failure(self):
        """
        Assert that failing to assign maintainer is handled correctly.
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.patch.side_effect = response_mock

        namespace = "namespace"
        repo = "repo"
        maintainer_fas = "amedvede"
        payload = {
            "EPEL Maintainer name": maintainer_fas,
            "Maintainer name": maintainer_fas,
        }

        expected_error = "Couldn't set new maintainer on project '{0}/{1}'".format(
            namespace, repo
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.assign_maintainer_to_project(namespace, repo, maintainer_fas)

        self.pagure._requests_session.patch.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/" + namespace + "/" + repo,
            data=json.dumps(payload),
            headers=self.pagure.get_auth_header(),
        )

    def test_assign_orphan_to_project(self, monkeypatch):
        """
        Assigning the orphan user as a maintainer should invalidate the cache
        """
        response_mock = Mock()
        response_mock.status_code = 200

        self.pagure._requests_session.patch.return_value = response_mock
        monkeypatch.setattr(cache, "delete", Mock(), raising=True)

        namespace = "namespace"
        repo = "repo"

        self.pagure.assign_maintainer_to_project(namespace, repo, "orphan")
        self.pagure._requests_session.patch.assert_called()
        cache.delete.assert_called_once_with(
            "toddlers.utils.pagure:is_project_orphaned|namespace repo"
        )

    def test_assign_orphan_to_project_failure(self, monkeypatch):
        """
        Failure to assign the orphan user as a maintainer should not invalidate the cache
        """
        response_mock = Mock()
        response_mock.status_code = 500

        self.pagure._requests_session.patch.side_effect = response_mock
        monkeypatch.setattr(cache, "delete", Mock(), raising=True)

        with pytest.raises(PagureError):
            self.pagure.assign_maintainer_to_project("namespace", "repo", "orphan")
        cache.delete.assert_not_called()


class TestPagureAddMemberToGroup:
    """
    Test class for `toddlers.pagure.Pagure.add_member_to_group` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_add_member_to_group(self):
        """
        Assert that adding member to group is processed correctly.
        """
        response_mock = Mock()

        self.pagure._requests_session.post.return_value = response_mock

        user = "conrad_kurze"
        group = "adeptus_astartes"
        payload = {
            "user": user,
        }

        self.pagure.add_member_to_group(user, group)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/group/" + group + "/add",
            data=json.dumps(payload),
            headers=self.pagure.get_auth_header(),
        )

    def test_add_member_to_group_failure(self):
        """
        Assert that failing to add member to group is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.post.return_value = response_mock

        user = "conrad_kurze"
        group = "adeptus_astartes"
        payload = {
            "user": user,
        }

        expected_error = "Couldn't add user '{0}' to group '{1}'".format(user, group)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.add_member_to_group(user, group)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/group/" + group + "/add",
            data=json.dumps(payload),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureRemoveMemberFromGroup:
    """
    Test class for `toddlers.pagure.Pagure.remove_member_from_group` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_remove_member_from_group(self):
        """
        Assert that removing member from group is processed correctly.
        """
        response_mock = Mock()

        self.pagure._requests_session.post.return_value = response_mock

        user = "conrad_kurze"
        group = "adeptus_astartes"
        payload = {
            "user": user,
        }

        self.pagure.remove_member_from_group(user, group)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/group/" + group + "/remove",
            data=json.dumps(payload),
            headers=self.pagure.get_auth_header(),
        )

    def test_remove_member_from_group_failure(self):
        """
        Assert that failing to remove member from group is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.post.return_value = response_mock

        user = "conrad_kurze"
        group = "adeptus_astartes"
        payload = {
            "user": user,
        }

        expected_error = "Couldn't remove user '{0}' from group '{1}'".format(
            user, group
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.remove_member_from_group(user, group)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/group/" + group + "/remove",
            data=json.dumps(payload),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetGroupMembers:
    """
    Test class for `toddlers.pagure.Pagure.get_group_members` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_group_members(self):
        """
        Assert that retrieving members in group is processed correctly.
        """
        response_mock = Mock()
        user = "conrad_kurze"
        data = {"members": [user]}
        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        group = "adeptus_astartes"

        result = self.pagure.get_group_members(group)

        assert result == [user]

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/group/" + group,
            headers=self.pagure.get_auth_header(),
        )

    def test_get_group_members_no_group(self):
        """
        Assert that when group doesn't exist is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False
        response_mock.status_code = 404

        self.pagure._requests_session.get.return_value = response_mock

        group = "adeptus_astartes"

        result = self.pagure.get_group_members(group)

        assert not result

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/group/" + group,
            headers=self.pagure.get_auth_header(),
        )

    def test_get_group_members_failure(self):
        """
        Assert that failing to obtain members of group is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.get.return_value = response_mock

        group = "adeptus_astartes"

        expected_error = "Couldn't get members of group '{0}'".format(group)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_group_members(group)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/group/" + group,
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetRetiredPackages:
    """
    Test class for `toddlers.utils.pagure.Pagure.get_retired_packages` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_retired_packages(self):
        """
        Assert that retrieving retired packages is processed correctly.
        """
        response_mock = Mock()
        branch = "f40"
        package_list = ["package01", "package02"]
        data = {branch: package_list}
        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        result = self.pagure.get_retired_packages(branch)

        assert result == package_list

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/lookaside/retired_in_{0}.json".format(branch),
            headers=self.pagure.get_auth_header(),
        )

    def test_get_retired_packages_malformed_data(self):
        """
        Assert that malformed data in response are handled correctly.
        """
        response_mock = MagicMock()
        branch = "f40"
        data = {}
        response_mock.json.return_value = data

        self.pagure._requests_session.get.return_value = response_mock

        expected_error = "Expected key not found in pagure response."

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_retired_packages(branch)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/lookaside/retired_in_{0}.json".format(branch),
            headers=self.pagure.get_auth_header(),
        )

    def test_get_retired_404(self):
        """
        Assert that 404 is correctly handled. 404 means that the branch doesn't exist.
        """
        response_mock = Mock()
        response_mock.ok = False
        response_mock.status_code = 404
        branch = "f40"

        self.pagure._requests_session.get.return_value = response_mock

        result = self.pagure.get_retired_packages(branch)

        assert result == []

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/lookaside/retired_in_{0}.json".format(branch),
            headers=self.pagure.get_auth_header(),
        )

    def test_get_retired_packages_failure(self):
        """
        Assert that failing to obtain members of group is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False
        branch = "f40"

        self.pagure._requests_session.get.return_value = response_mock

        expected_error = "Couldn't retrieve retired packages for '{0}'".format(branch)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_retired_packages(branch)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/lookaside/retired_in_{0}.json".format(branch),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureIsRetiredOnBranch:
    """
    Test class for `toddlers.utils.pagure.Pagure.is_retired_on_branch` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure.get_retired_packages = Mock()

    def test_is_retired_on_branch_true(self):
        """
        Assert that retired package is correctly processed.
        """
        # Prepare
        branch = "f40"
        package_list = ["package01", "package02"]
        package = "package01"

        self.pagure.get_retired_packages.return_value = package_list

        # Run test
        result = self.pagure.is_retired_on_branch(package, branch)

        # Asserts
        assert result is True

        self.pagure.get_retired_packages.assert_called_with(branch)

    def test_is_retired_on_branch_false(self):
        """
        Assert that non-retired package is correctly processed.
        """
        # Prepare
        branch = "f40"
        package_list = ["package02"]
        package = "package01"

        self.pagure.get_retired_packages.return_value = package_list

        # Run test
        result = self.pagure.is_retired_on_branch(package, branch)

        # Asserts
        assert result is False

        self.pagure.get_retired_packages.assert_called_with(branch)


class TestPagureHasDeadPackageOnBranch:
    """
    Test class for `toddlers.utils.pagure.Pagure.has_dead_package_on_branch` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://src.stg.fedoraproject.org",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_is_retired_on_branch_true(self):
        """
        Assert that retired package is correctly processed.
        """
        # Prepare
        branch = "f40"
        package = "package01"

        response_mock = MagicMock()
        response_mock.ok = True
        response_mock.json.return_value = {
            "content_url": "https://src.stg.fedoraproject.org/api",
            "name": "dead.package",
            "type": "file",
        }

        self.pagure._requests_session.get.return_value = response_mock

        # Run test
        result = self.pagure.has_dead_package_on_branch(package, branch)

        # Asserts
        assert result is True

    def test_is_retired_on_branch_false(self):
        """
        Assert that non-retired package is correctly processed.
        """
        # Prepare
        branch = "f40"
        package = "package01"

        response_mock = MagicMock()
        response_mock.ok = False
        response_mock.json.return_value = {
            "error": "File not found in this git repository",
            "error_code": "EFILENOTFOUND",
        }
        response_mock.status_code = 404

        self.pagure._requests_session.get.return_value = response_mock

        # Run test
        result = self.pagure.has_dead_package_on_branch(package, branch)

        # Asserts
        assert result is False

    def test_is_retired_on_branch_error(self):
        """
        Assert that non-retired package is correctly processed.
        """
        # Prepare
        branch = "f40"
        package = "package01"

        response_mock = MagicMock()
        response_mock.ok = False
        response_mock.json.return_value = {
            "error": "File not found in this git repository",
            "error_code": "EFILENOTFOUND",
        }
        response_mock.status_code = 501

        self.pagure._requests_session.get.return_value = response_mock

        # Run test
        result = self.pagure.has_dead_package_on_branch(package, branch)

        # Asserts
        assert result is False


class TestPagureGetAllGroups:
    """
    Test class for `toddlers.utils.pagure.Pagure.get_all_groups` method.
    """

    @pytest.mark.parametrize(
        "url,message",
        (
            (
                "https://stg.pagure.io",
                "Retrieving groups from https://stg.pagure.io/api/0/groups",
            ),
            (
                "https://stg.src.fedoraproject.org",
                "Retrieving groups from https://stg.src.fedoraproject.org/api/0/groups",
            ),
        ),
        ids=(
            "pagure_url",
            "distgit_url",
        ),
    )
    def test_get_all_groups(self, url, message, caplog):
        """
        Assert correct retrieval of distgit groups
        """
        config = {
            "pagure_url": url,
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()
        caplog.set_level(logging.INFO)
        response_mock = MagicMock()
        response_mock.ok.return_value = True
        response_mock.json.return_value = {
            "groups": [
                "group_a",
                "group_b",
            ],
            "pagination": {
                "next": None,
            },
        }

        # mock_request.side_effect = request
        self.pagure._requests_session.get.return_value = response_mock

        result = self.pagure.get_all_groups()
        assert result == [
            "group_a",
            "group_b",
        ]
        assert caplog.records[-1].message == message


class TestPagureOrphanPackage:
    """
    Test class for `toddlers.utils.pagure.Pagure.orphan_package` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_orphan_package(self, monkeypatch):
        """
        Assert that orphaning package is processed correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = True

        self.pagure._requests_session.post.return_value = response_mock

        monkeypatch.setattr(cache, "delete", Mock(), raising=True)

        namespace = "rpms"
        package = "foo"
        reason = "reason"
        info = "info"

        self.pagure.orphan_package(namespace, package, reason, info)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/_dg/orphan/{0}/{1}".format(namespace, package),
            data=json.dumps({"orphan_reason": reason, "orphan_reason_info": info}),
            headers=self.pagure.get_auth_header(),
        )
        cache.delete.assert_called_once_with(
            "toddlers.utils.pagure:is_project_orphaned|rpms foo"
        )

    def test_orphan_package_failure(self, monkeypatch):
        """
        Assert that failing to orphan package is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.post.return_value = response_mock

        monkeypatch.setattr(cache, "delete", Mock(), raising=True)

        namespace = "rpms"
        package = "foo"
        reason = "reason"
        info = "info"

        expected_error = "Couldn't orphan package '{0}/{1}'".format(namespace, package)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.orphan_package(namespace, package, reason, info)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/_dg/orphan/{0}/{1}".format(namespace, package),
            data=json.dumps({"orphan_reason": reason, "orphan_reason_info": info}),
            headers=self.pagure.get_auth_header(),
        )
        cache.delete.assert_not_called()


class TestSetBugzillaOverrides:
    """
    Test class for `toddlers.utils.pagure.Pagure.set_bugzilla_overrides` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_set_bugzilla_overrides(self):
        """
        Assert that setting bugzilla maintainers is processed correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = True

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        fedora_assignee = "Perturabo"
        epel_assignee = "Angron"

        self.pagure.set_bugzilla_overrides(
            namespace, package, fedora_assignee, epel_assignee
        )

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/{0}/{1}".format(namespace, package),
            data=json.dumps(
                {"fedora_assignee": fedora_assignee, "epel_assignee": epel_assignee}
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_set_bugzilla_overrides_failure(self):
        """
        Assert that failing to set bugzilla maintainers is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        fedora_assignee = "Perturabo"
        epel_assignee = "Angron"

        expected_error = "Couldn't set bugzilla maintainers for '{0}/{1}'".format(
            namespace, package
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.set_bugzilla_overrides(
                namespace, package, fedora_assignee, epel_assignee
            )

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/_dg/bzoverrides/{0}/{1}".format(namespace, package),
            data=json.dumps(
                {"fedora_assignee": fedora_assignee, "epel_assignee": epel_assignee}
            ),
            headers=self.pagure.get_auth_header(),
        )

    def test_set_bugzilla_overrides_orphan(self, monkeypatch):
        """
        Setting the orphan user as bugzilla maintainer should invalidate the cache.
        """
        response_mock = MagicMock()
        response_mock.ok = True

        self.pagure._requests_session.post.return_value = response_mock
        monkeypatch.setattr(cache, "delete", Mock(), raising=True)

        namespace = "rpms"
        package = "foo"

        for distro in ("fedora", "epel"):
            kwargs = {f"{distro}_assignee": "orphan"}
            self.pagure.set_bugzilla_overrides(namespace, package, **kwargs)

        self.pagure._requests_session.post.call_count == 2
        assert cache.delete.call_count == 2
        cache.delete.assert_called_with(
            "toddlers.utils.pagure:is_project_orphaned|rpms foo"
        )


class TestModifyACLs:
    """
    Test class for `toddlers.utils.pagure.Pagure.modify_acls` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_modify_acls(self):
        """
        Assert that modifying ACLs is processed correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = True

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        user_type = "user"
        name = "Angron"

        self.pagure.modify_acls(namespace, package, user_type, name)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/modifyacls".format(namespace, package),
            data=json.dumps({"user_type": "user", "name": name}),
            headers=self.pagure.get_auth_header(),
        )

    def test_modify_acls_failure(self):
        """
        Assert that failing to modify ACLs is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        user_type = "user"
        name = "Angron"

        expected_error = "Couldn't modify ACLs for '{0}/{1}'".format(namespace, package)

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.modify_acls(namespace, package, user_type, name)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/git/modifyacls".format(namespace, package),
            data=json.dumps({"user_type": "user", "name": name}),
            headers=self.pagure.get_auth_header(),
        )


class TestUpdateWatcherStatus:
    """
    Test class for `toddlers.utils.pagure.Pagure.update_watcher_status` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_modify_acls(self):
        """
        Assert that updating watcher status is processed correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = True

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        status = -1
        watcher = "Emperor of Mankind"

        self.pagure.update_watcher_status(namespace, package, status, watcher)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/watchers/update".format(
                namespace, package
            ),
            data=json.dumps({"status": status, "watcher": watcher}),
            headers=self.pagure.get_auth_header(),
        )

    def test_modify_acls_failure(self):
        """
        Assert that failing to modify ACLs is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        status = -1
        watcher = "Emperor of Mankind"

        expected_error = "Couldn't update watcher status on '{0}/{1}'".format(
            namespace, package
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.update_watcher_status(namespace, package, status, watcher)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/watchers/update".format(
                namespace, package
            ),
            data=json.dumps({"status": status, "watcher": watcher}),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureGetPlugins:
    """
    Test class for `toddlers.utils.pagure.Pagure.get_plugins` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_get_plugins(self):
        """
        Assert correct retrieval of plugins enabled on repository
        """
        namespace = "rpms"
        package = "foo"

        response_mock = MagicMock()
        response_mock.ok.return_value = True
        response_mock.json.return_value = {
            "plugins": [
                {"plugin_01": {}},
                {"plugin_02": {}},
            ]
        }

        # mock_request.side_effect = request
        self.pagure._requests_session.get.return_value = response_mock

        result = self.pagure.get_plugins(namespace, package)
        assert result == [
            "plugin_01",
            "plugin_02",
        ]

    def test_get_plugins_failure(self):
        """
        Assert that failing to get plugins is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False
        response_mock.headers = {"content-type": "application/json"}

        self.pagure._requests_session.get.return_value = response_mock

        namespace = "rpms"
        package = "foo"

        expected_error = "Failed to get enabled plugins on '{0}/{1}'".format(
            namespace, package
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.get_plugins(namespace, package)

        self.pagure._requests_session.get.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/settings/plugins".format(
                namespace, package
            ),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureDisablePlugin:
    """
    Test class for `toddlers.utils.pagure.Pagure.disable_plugin` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_disable_plugin(self):
        """
        Assert plugin is correctly disabled on repository
        """
        namespace = "rpms"
        package = "foo"
        plugin = "Some plugin"

        response_mock = MagicMock()
        response_mock.ok.return_value = True

        # mock_request.side_effect = request
        self.pagure._requests_session.post.return_value = response_mock

        self.pagure.disable_plugin(namespace, package, plugin)
        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/settings/{2}/remove".format(
                namespace, package, urllib.parse.quote(plugin)
            ),
            data=json.dumps({}),
            headers=self.pagure.get_auth_header(),
        )

    def test_disable_plugin_failure(self):
        """
        Assert that failing to disable plugin is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False
        response_mock.headers = {"content-type": "application/json"}

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        plugin = "Some plugin"

        expected_error = "Failed to disable plugin on '{0}/{1}'".format(
            namespace, package
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.disable_plugin(namespace, package, plugin)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/settings/{2}/remove".format(
                namespace, package, urllib.parse.quote(plugin)
            ),
            data=json.dumps({}),
            headers=self.pagure.get_auth_header(),
        )


class TestPagureEnablePlugin:
    """
    Test class for `toddlers.utils.pagure.Pagure.enable_plugin` method.
    """

    def setup_method(self):
        """
        Setup method for the test class.
        """
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "Very secret key",
        }
        self.pagure = pagure.set_pagure(config)
        self.pagure._requests_session = Mock()

    def test_enable_plugin(self):
        """
        Assert plugin is correctly enabled on repository
        """
        namespace = "rpms"
        package = "foo"
        plugin = "Some plugin"

        response_mock = MagicMock()
        response_mock.ok.return_value = True

        # mock_request.side_effect = request
        self.pagure._requests_session.post.return_value = response_mock

        self.pagure.enable_plugin(namespace, package, plugin)
        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/settings/{2}/install".format(
                namespace, package, urllib.parse.quote(plugin)
            ),
            data=json.dumps({}),
            headers=self.pagure.get_auth_header(),
        )

    def test_enable_plugin_failure(self):
        """
        Assert that failing to enable plugin is handled correctly.
        """
        response_mock = MagicMock()
        response_mock.ok = False
        response_mock.headers = {"content-type": "application/json"}

        self.pagure._requests_session.post.return_value = response_mock

        namespace = "rpms"
        package = "foo"
        plugin = "Some plugin"

        expected_error = "Failed to enable plugin on '{0}/{1}'".format(
            namespace, package
        )

        with pytest.raises(PagureError, match=expected_error):
            self.pagure.enable_plugin(namespace, package, plugin)

        self.pagure._requests_session.post.assert_called_with(
            "https://pagure.io/api/0/{0}/{1}/settings/{2}/install".format(
                namespace, package, urllib.parse.quote(plugin)
            ),
            data=json.dumps({}),
            headers=self.pagure.get_auth_header(),
        )
