import abc
import logging

import fedora_messaging.config
import fedora_messaging.exceptions

from .utils.cache import set_cache
from .utils.misc import merge_dicts


_log = logging.getLogger(__name__)


def _get_toddler_config(toddler_name: str) -> dict:
    """Merge default and private configuration for a toddler."""
    base_config = fedora_messaging.config.conf["consumer_config"]

    default_config = base_config.get("default", {})
    private_config = base_config.get(toddler_name, {})

    return merge_dicts(default_config, private_config)


class ToddlerBase(metaclass=abc.ABCMeta):

    def __init__(self):
        self.reset_routing_keys()
        self._config = _get_toddler_config(self.name)
        set_cache(fedora_messaging.config.conf["consumer_config"].get("cache", {}))

    def reset_routing_keys(self):
        # Only for dev/debug: in Fedora infra all bindings are set in ansible
        for binding in fedora_messaging.config.conf["bindings"]:
            if binding["routing_keys"] == ["#"]:
                _log.info(
                    "Updating the routing_keys of the queue %s to the topics of interest",
                    binding["queue"],
                )
                binding["routing_keys"] = self.amqp_topics

    @property
    @abc.abstractmethod
    def name(self):
        """Returns name of the plugin."""
        return "base"

    @property
    @abc.abstractmethod
    def amqp_topics(self):
        """Returns the list of topics of interest for this toddler in a format
        that can be used directly when connecting to amqp.
        For example, it supports items like:
            ``org.fedoraproject.#.buildsys.build.state.change``
        which is valid when subscribing to a queue in amqp but will not allow
        string based comparison with the topic extracted from the message.

        """
        return []

    @abc.abstractmethod
    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return

    @abc.abstractmethod
    def process(self, config, message):
        """Process a given message."""
        return

    def __call__(self, message):
        try:
            return self.process(self._config, message)
        except Exception:
            _log.exception(
                "Toddler '%s' failed to process message id: %s -- putting it back in the queue",
                self.name,
                message.id,
            )
            raise fedora_messaging.exceptions.Nack("A toddler tripped")
