"""
This script queries Bodhi and dist_git for all the packages in Fedora, it then go through
each of them and determines if they are inactive on all branches or not.
If they are, it will then query dist-git to retrieve all of its current
maintainers, if they do not have ``orphan`` as main user they will be orphaned
and all co-maintainers will be removed.

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import logging

from toddlers.base import ToddlerBase
from toddlers.exceptions import PagureError
from toddlers.utils import bodhi, notify, pagure
from toddlers.utils.requests import make_session


_log = logging.getLogger(__name__)


class CleanRetiredPackages(ToddlerBase):
    """
    Listens to messages sent by playtime (which lives in toddlers) and check
    if the `email_overrides.toml` file stored in ansible is up to date.
    """

    name = "clean_retired_packages"

    amqp_topics = [
        "org.fedoraproject.*.toddlers.trigger.clean_retired_packages",
    ]

    namespace = "rpms"

    def __init__(self):
        super().__init__()
        self.requests_session = make_session()
        self.dist_git = None
        self.bodhi = None

    def accepts_topic(self, topic):
        """
        Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "toddlers.trigger.clean_retired_packages"
        )

    def process(self, config, message, dry_run=False):
        """Process a given message."""
        self.dist_git = pagure.set_pagure(
            {
                "pagure_url": config.get("dist_git_url"),
                "pagure_api_key": config.get("dist_git_token"),
            }
        )

        self.bodhi = bodhi.set_bodhi(config)

        _log.info("Getting active branches from Bodhi")
        active_branches = self.bodhi.get_active_branches()

        _log.info("Retrieving list of packages retired in dist_git")
        retired_packages = dict()
        for branch in active_branches:
            for package in self.dist_git.get_retired_packages(branch):
                if package not in retired_packages:
                    retired_packages[package] = []
                retired_packages[package].append(branch)

        # Check for fully retired packages
        fully_retired = []
        for package in retired_packages.keys():
            try:
                branches = self.dist_git.get_branches(self.namespace, package)
            except PagureError as e:
                _log.info(str(e) + " Skipping.")
                continue
            branches.sort()
            retired_packages[package].sort()
            if retired_packages[package] == branches:
                fully_retired.append(package)

        _log.info(
            "%s fully retired packages found in %s", len(fully_retired), self.namespace
        )

        logs = []

        _log.info("Processing packages")
        for package in fully_retired:
            _log.info("    %s", package)

            try:
                data = self.dist_git.get_project(self.namespace, package)
            except PagureError:
                _log.info(
                    "   %s returned error -- skipping", self.namespace + "/" + package
                )
                continue

            if data["user"]["name"] != "orphan":
                log = "Orphaning %s from %s" % (
                    self.namespace + "/" + package,
                    data["user"]["name"],
                )
                _log.info(log)
                logs.append(log)
                # Orphan packages
                try:
                    self.dist_git.orphan_package(
                        self.namespace,
                        package,
                        "Other",
                        "Package was retired before it was orphaned",
                    )
                    self.dist_git.set_bugzilla_overrides(
                        self.namespace, package, None, None
                    )
                except PagureError:
                    continue

            packagers = set()
            for lvl in data["access_users"]:
                packagers.update([u for u in data["access_users"][lvl]])

            for packager in sorted(packagers):
                if packager == "orphan":
                    continue
                log = "Removing access on %s/%s to %s" % (
                    self.namespace,
                    package,
                    packager,
                )
                _log.info(log)
                logs.append(log)
                # Remove access
                try:
                    self.dist_git.modify_acls(
                        self.namespace,
                        package,
                        "user",
                        packager,
                    )
                    self.dist_git.update_watcher_status(
                        self.namespace, package, -1, packager
                    )
                except PagureError:
                    continue

            groups = set()
            for lvl in data["access_groups"]:
                groups.update([u for u in data["access_groups"][lvl]])

            for group in sorted(groups):
                log = "Removing access on %s/%s to @%s" % (
                    self.namespace,
                    package,
                    group,
                )
                _log.info(log)
                logs.append(log)
                # Remove access
                try:
                    self.dist_git.modify_acls(
                        self.namespace,
                        package,
                        "group",
                        group,
                    )
                except PagureError:
                    continue

        if logs:
            logs_text = "\n- ".join(logs)
            notify.notify_admins_on_retired_packages_cleanup(
                to_addresses=[config["admin_email"]],
                from_address=config["admin_email"],
                mail_server=config["mail_server"],
                logs_text=logs_text,
            )
