"""
When a branch is retired, automatically blocks the package on Koji

Authors:    Anton Medvedev <amedvede@redhat.com>
"""

from collections import defaultdict
import logging
import re
import time

import koji
import requests

from toddlers.base import ToddlerBase
from toddlers.utils import bodhi, pagure


_log = logging.getLogger(__name__)


class KojiBlockRetired(ToddlerBase):
    """Listens to messages sent by playtime (which lives in toddlers) to sync
    all the packagers found in FAS to bugzilla.
    """

    name = "koji_block_retired"

    amqp_topics = [
        "org.fedoraproject.*.git.receive",
        "org.fedoraproject.*.toddlers.trigger.koji_block_retired",
    ]

    def __init__(self):
        super().__init__()
        self.koji_session = None
        self.koji_url = None
        self.koji_principal = None
        self.koji_keytab = None
        self.bodhi = None
        self.distgit = None

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        if topic.startswith("org.fedoraproject."):
            if topic.endswith("git.receive"):
                return True
            if topic.endswith("toddlers.trigger.koji_block_retired"):
                return True
        return False

    def _create_session(self):  # pragma: no cover
        """Makes a koji session, that handles logging in"""
        self.koji_session = koji.ClientSession(self.koji_url)
        self.koji_session.gssapi_login(
            principal=self.koji_principal, keytab=self.koji_keytab
        )

    def get_tag_from_target(self, branch):
        try:
            _log.info(f"Getting koji tag for {branch} target")
            target = self.koji_session.getBuildTarget(branch)
            tag = target["build_tag_name"].removesuffix("-build")
            return tag
        except koji.GenericError as e:
            _log.exception(f"Failed to get {branch} tag in koji: {e}")
        return None

    def adjust_releases_for_lookaside(self, releases):
        releases = list(set(releases))  # Let's remove duplicates
        if any(re.match(r"^f\d{2,}$", regex) for regex in releases):
            if "rawhide" in releases:
                releases.remove("rawhide")
            # Remove the highest value, which should be the rawhide
            releases.remove(max(releases))
            # And add rawhide instead
            releases.append("rawhide")
        lookaside_release = []
        for release in releases:
            if "eln" in release or "-next" in release:
                continue
            lookaside_release.append(release)
        lookaside_release.sort()
        return lookaside_release

    def find_and_block(self, config):
        self.bodhi = bodhi.set_bodhi(config)
        active_releases = self.bodhi.get_active_branches()
        adjusted = self.adjust_releases_for_lookaside(active_releases)
        needs_blocking = defaultdict(list)
        for active_release in adjusted:
            # Do not block on frozen branches
            if self.bodhi.is_branch_frozen(active_release):
                _log.info(f"Branch {active_release} is frozen, ignoring.")
                continue
            # Do not block on fedora stable branches (have EOL)
            if re.match(r"^f", active_release):
                if self.bodhi.is_branch_stable(active_release):
                    _log.info(f"Cannot block on stable branches: {active_release}")
                    continue
            retired_url = f"{config['lookaside_url']}/retired_in_{active_release}.json"
            response = requests.get(retired_url)
            if not response.ok:
                _log.error(
                    "Not able to get data from the lookaside cache "
                    f"{retired_url}: {response.status_code}, {response.text}"
                )
                return
            packages = response.json()[active_release]

            for package in packages:
                if active_release == "rawhide":
                    active_release = self.get_tag_from_target("rawhide")
                elif "epel" in active_release:
                    active_release = self.get_tag_from_target(active_release)
                _log.info(
                    f"Checking if package {package} is blocked "
                    f"in koji for release {active_release}"
                )
                try:
                    (resp,) = self.execute_koji_call(
                        "listPackages",
                        pkgID=package,
                        tagID=active_release,
                    )
                except TypeError:
                    _log.error("Not able to get data from koji, bailing.")
                    return

                if not resp["blocked"]:
                    _log.info(f"Package {package} is not blocked in koji.")

                    _log.info(
                        f"Last check before blocking: is package {package} "
                        f"really retired on {active_release}?"
                    )
                    self.dist_git = pagure.set_pagure(
                        {
                            "pagure_url": config.get("dist_git_url"),
                            "pagure_api_key": config.get("dist_git_token"),
                        }
                    )
                    if not self.dist_git.has_dead_package_on_branch(
                        package, active_release
                    ):
                        _log.info(
                            f"Package {package} is not retired on the distgit branch "
                            f"{active_release}, was probably unretired recently, bailing."
                        )
                        return

                    _log.info(f"Blocking package {package}, tag: {active_release}")
                    needs_blocking[active_release].append(package)
                    self.execute_koji_call(
                        "packageListBlock",
                        taginfo=active_release,
                        pkginfo=package,
                    )
                else:
                    _log.info(
                        f"Package {package} is already blocked in koji "
                        f"for tag {active_release}"
                    )
        _log.info(
            f"All packages that should be blocked in this run: {dict(needs_blocking)}"
        )

    def process(self, config, message):
        """
        Blocks the package on Koji

        This step will check if commit has dead.package file added and block package on Koji.
        """
        topic = message.topic
        msg = message.body

        self.config = config

        self.koji_url = f"{self.config['koji_url']}/kojihub"
        self.principal = self.config["principal"]
        self.keytab = self.config["keytab"]

        if topic.endswith("toddlers.trigger.koji_block_retired"):
            if self.koji_session is None:
                self._create_session()  # pragma: no cover
            self.find_and_block(config)

        if topic.endswith("git.receive"):
            # If there is no dead.package file in commit, then it can be ignored
            if "dead.package" not in msg["commit"]["stats"]["files"]:
                _log.info("No dead.package in the commit, bailing")
                return

            dead_package_commit = msg["commit"]["stats"]["files"]["dead.package"]
            if (
                dead_package_commit["additions"] == 0
                and dead_package_commit["deletions"] != 0
            ):
                _log.info("dead.package file was not added, bailing")
                return

            repo = msg["commit"]["repo"]
            branch = msg["commit"]["branch"]

            # Filter out cases when a new branch with a history of retiring is created.
            # In that case the commit with a dead.package will be re-sent and would make the toddler
            # do the blocking even when not desired.
            # Check if the package is really retired on distgit.
            self.dist_git = pagure.set_pagure(
                {
                    "pagure_url": config.get("dist_git_url"),
                    "pagure_api_key": config.get("dist_git_token"),
                }
            )
            if not self.dist_git.has_dead_package_on_branch(repo, branch):
                _log.info(
                    f"Package {repo} is not retired on the distgit branch "
                    f"{branch}, probably new branch creation, bailing"
                )
                return

            # Exclude frozen branches from blocking
            self.bodhi = bodhi.set_bodhi(config)
            if self.bodhi.is_branch_frozen(branch):
                _log.info(
                    f"Cannot block package {repo} on frozen branch {branch}, bailing."
                )
                return

            # Retirements are allowed on rawhide, branched and EPEL
            # and so is koji blocking.
            # Exclude stable fedora branches from blocking.
            # Stable branches are defined as those that have set EOL date.
            if re.match(r"^f", branch):
                if self.bodhi.is_branch_stable(branch):
                    _log.info(f"Cannot block on stable branches: {branch}")
                    return

            if self.koji_session is None:
                self._create_session()  # pragma: no cover

            if branch in ["main", "rawhide"]:
                tag_name = self.get_tag_from_target("rawhide")
            elif "container" in msg["commit"]["namespace"]:
                tag_name = (
                    "".join([branch, "-container"])
                    if "container" not in branch
                    else branch
                )
            elif re.match(r"^epel\d{2,}$", branch):
                tag_name = self.get_tag_from_target(branch)
            else:
                tag_name = branch

            _log.info("Processing Koji block retired for %s", repo)

            _log.info(f"Blocking package {repo}, tag: {tag_name}")
            self.execute_koji_call(
                "packageListBlock",
                taginfo=tag_name,
                pkginfo=repo,
            )

    def execute_koji_call(self, action, num_of_attempts=2, **kwargs):
        raw_data = None
        for i in range(num_of_attempts):
            try:
                raw_data = getattr(self.koji_session, action)(**kwargs)
                break
            except (koji.GenericError, koji.ActionNotAllowed) as e:
                if i >= num_of_attempts - 1:
                    _log.exception(
                        f"Unable to create koji ClientSession for call {action}: {e}"
                    )
                _log.error(f"{e}, recreating koji_session and trying again")
                self._create_session()
                # Wait a bit and try again
                time.sleep(5)
        return raw_data
