"""
This script takes as input the fedmsg messages published under the topic
``toddlers.trigger.packager_bugzilla_sync`` and runs a sync of all packagers
found in FAS to bugzilla so that they can do thins like editing flags on Fedora
tickets.

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import argparse
import logging
import sys

import tomllib

try:
    import tqdm
except ImportError:
    tqdm = None

from ..base import ToddlerBase
from ..utils import bugzilla_system
from ..utils import fedora_account


_log = logging.getLogger(__name__)


class PackagerBugzillaSync(ToddlerBase):
    """Listens to messages sent by playtime (which lives in toddlers) to sync
    all the packagers found in FAS to bugzilla.
    """

    name = "packager_bugzilla_sync"

    amqp_topics = [
        "org.fedoraproject.*.toddlers.trigger.packager_bugzilla_sync",
    ]

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "toddlers.trigger.packager_bugzilla_sync"
        )

    def process(self, config, message, username=False, dry_run=False):
        """Process a given message."""

        try:
            with open(config["email_overrides_file"], "rb") as overrides_fh:
                email_overrides = tomllib.load(overrides_fh)
        except Exception:
            print("Failed to load the file containing the email-overrides")
            raise

        _log.info("Setting up connection to FAS")
        fedora_account.set_fasjson(config)

        if not username:
            # Retrieve all the packagers in FAS:
            _log.info("Retrieving the list of packagers in FAS")
            fas_packagers = sorted(fedora_account.get_group_member("packager"))
        else:
            fas_packagers = [username]

        n_packagers = len(fas_packagers)
        _log.info(f"{n_packagers} packagers found in FAS")
        fas_packagers_info = {}
        _log.info("Retrieving the bugzilla email for each packager")

        # If the import fails, no progress bar
        # At DEBUG or below, we're showing things at each iteration so the progress
        # bar doesn't look good.
        # At WARNING or above, we do not want to show anything.
        if (
            tqdm is not None and _log.getEffectiveLevel() == logging.INFO
        ):  # pragma no cover
            fas_packagers = tqdm.tqdm(fas_packagers)

        for idx, username in enumerate(fas_packagers):
            _log.debug(
                f"   Retrieving bz email of user {username}: {idx}/{n_packagers}"
            )
            bz_email = fedora_account.get_bz_email_user(username, email_overrides)
            fas_packagers_info[bz_email] = username

        _log.info("Setting up connection to bugzilla")
        bugzilla_system.set_bz(config)

        # Retrieve all the packagers in bugzilla
        _log.info("Retrieving the list of packagers in bugzilla")
        bz_packagers = bugzilla_system.get_group_member(config["bugzilla_group"])
        n_bz_packagers = len(bz_packagers)
        _log.info(
            f"{n_bz_packagers} members of {config['bugzilla_group']} found in bugzilla"
        )

        fas_set = set(fas_packagers_info)
        bz_set = set(bz_packagers)

        overlap = len(fas_set.intersection(bz_set))
        fas_only = fas_set - bz_set
        bz_only = bz_set - fas_set
        # To prevent TypeError remove any None element from set
        fas_only.discard(None)

        _log.info(f"{overlap} packagers found in both places")
        _log.info(f"{len(fas_only)} packagers found only in FAS (to be added)")
        _log.info(f"{len(bz_only)} packagers found only in BZ (to be removed)")

        # Store a list of user with no bugzilla account
        no_bz_account = []
        # Add the packagers found only in FAS to the bugzilla group
        _log.info(f"Adding to {config['bugzilla_group']} the packagers found in FAS")
        for user_email in sorted(fas_only):
            no_bz_account = bugzilla_system.add_user_to_group(
                user_email=user_email,
                bz_group=config["bugzilla_group"],
                no_bz_account=no_bz_account,
                dry_run=dry_run,
            )

        _log.info(f"{len(no_bz_account)} emails had no corresponding bugzilla account")


# We have had the situation in the past where we've had to check a specific
# account, so the following code allows to run this script stand-alone if
# needed.


def get_arguments(args):
    """Load and parse the CLI arguments."""
    parser = argparse.ArgumentParser(
        description="Sync packagers access from FAS to bugzilla"
    )
    parser.add_argument(
        "conf",
        help="Configuration file",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Do not change anything on bugzilla",
    )

    parser.add_argument(
        "username",
        default=None,
        nargs="?",
        help="Process a specific user instead of all the packagers",
    )

    log_level_group = parser.add_mutually_exclusive_group()
    log_level_group.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        dest="log_level",
        const=logging.WARNING,
        default=logging.INFO,
        help="Be less talkative",
    )
    log_level_group.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        help="Enable debugging output",
    )

    return parser.parse_args(args)


def setup_logging(log_level: int):
    handlers = []

    _log.setLevel(log_level)
    # We want all messages logged at level INFO or lower to be printed to stdout
    info_handler = logging.StreamHandler(stream=sys.stdout)
    handlers.append(info_handler)

    if log_level == logging.INFO:
        # In normal operation, don't decorate messages
        for handler in handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    logging.basicConfig(level=log_level, handlers=handlers)


def main(args):
    """Schedule the first test and run the scheduler."""
    args = get_arguments(args)
    setup_logging(log_level=args.log_level)

    with open(args.conf, "rb") as conf_fh:
        config = tomllib.load(conf_fh)
    PackagerBugzillaSync().process(
        config=config.get("consumer_config", {}).get("packager_bugzilla_sync", {}),
        message={},
        username=args.username,
        dry_run=args.dry_run,
    )


if __name__ == "__main__":  # pragma: no cover
    try:
        main(sys.argv[1:])
    except KeyboardInterrupt:
        pass
