"""
This script takes as input the fedora messages published under the topic
``pagure.issue.new`` and process the tickets on
https://pagure.io/releng/fedora-scm-requests/issues.

Authors:    Michal Konecny <mkonecny@redhat.com>

"""

import argparse
import fnmatch
import json
import logging
import re
import sys
from tempfile import TemporaryDirectory
import traceback
from typing import Optional

import arrow
from fedora_messaging.api import Message
from pagure_messages.issue_schema import IssueNewV1
import tomllib

from toddlers.base import ToddlerBase
from toddlers.exceptions import ValidationError
from toddlers.utils import bugzilla_system, fedora_account, git, pagure, requests

# Regex for branch name validation
STREAM_NAME_REGEX = r"^[a-zA-Z0-9.\-_+]+$"
# Regex for project name validation
PROJECT_NAME_REGEX = r"^[a-zA-Z0-9_][a-zA-Z0-9-_.+]*$"
# Regex for epel branch validation
EPEL_REGEX = r"^epel\d+(?:\.\d+)?(?:-next)?$"

# Plugin preventing branch creation
PLUGIN_NAME = "Prevent creating new branches by git push"

# Where to look for the scm-requests tickets
PROJECT_NAMESPACE = "releng/fedora-scm-requests"

# ERROR returned for invalid EPEL request
INVALID_EPEL_ERROR = (
    "This package is already an EL package and is built on all supported "
    "arches, therefore, it cannot be in EPEL. If this is a mistake or you "
    "have an exception, please contact the Release Engineering team at"
    "https://pagure.io/releng/issues"
)

_log = logging.getLogger(__name__)


class SCMRequestProcessor(ToddlerBase):
    """
    Listens for new tickets in https://pagure.io/releng/fedora-scm-requests/issues
    and process them, either by creating the branch or rejecting the ticket.
    """

    name: str = "scm_request_processor"

    amqp_topics: list = [
        "io.pagure.*.pagure.issue.new",
        "io.pagure.*.pagure.issue.comment.added",
    ]

    # SLAs for branches
    branch_slas: dict = {}

    # Monitoring choices for release-monitoring.org
    monitoring_choices: list = []

    # Mapping of pagure namespace to component
    pagure_namespace_to_component: dict = {}

    # Mapping of pagure namespace to product
    pagure_namespace_to_product: dict = {}

    # Pagure object connected to pagure.io
    pagure_io: pagure.Pagure

    # Pagure object connected to dist-git
    dist_git: pagure.Pagure

    # Path to temporary dir
    temp_dir: str = ""

    # Requests session
    requests_session: requests.requests.Session

    # Validation comment
    # This is the comment we will be looking for in the tickets
    # That need to be manually verified
    validation_comment: str = ""

    # Ping comment
    # Comment to sent when the ticket needs to be manually verified
    ping_comment: str = ""

    # Pagure user that will be creating the comments on pagure
    # for toddler
    pagure_user: str = ""

    def accepts_topic(self, topic: str) -> bool:
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.

        :arg topic: Topic to check.

        :returns: True if topic is accepted, False otherwise.
        """
        if topic.startswith("io.pagure."):
            if topic.endswith("pagure.issue.new"):
                return True
            if topic.endswith("pagure.issue.comment.added"):
                return True

        return False

    def process(
        self,
        config: dict,
        message: Message,
    ) -> None:
        """Process a given message.

        :arg config: Toddlers configuration
        :arg message: Message to process
        """
        _log.debug(
            "Processing message:\n{0}".format(json.dumps(message.body, indent=2))
        )
        project_name = message.body["project"]["fullname"]

        if project_name != PROJECT_NAMESPACE:
            _log.info(
                "The message doesn't belong to project {0}. Skipping message.".format(
                    PROJECT_NAMESPACE
                )
            )
            return

        issue = message.body["issue"]

        if issue["status"] != "Open":
            _log.info(
                "The issue {0} is not open. Skipping message.".format(issue["id"])
            )
            return

        self.branch_slas = config.get("branch_slas", {})
        self.monitoring_choices = config.get("monitor_choices", [])
        self.pagure_namespace_to_component = config.get(
            "pagure_namespace_to_component", {}
        )
        self.pagure_namespace_to_product = config.get("pagure_namespace_to_product", {})
        self.temp_dir = config.get("temp_folder", "")
        self.requests_session = requests.make_session()
        self.validation_comment = config.get("validation_comment", "")
        self.ping_comment = config.get("ping_comment", "")
        self.pagure_user = config.get("pagure_user", "")
        self.dist_git_token = config.get("dist_git_token", "")
        self.oidc_distgit_token = config.get("oidc_distgit_token", "")

        # Check if the message isn't generated by toddler user
        # Ignore the ticket if it is
        if message.body["agent"] == self.pagure_user:
            _log.info(
                "Last change on the ticket was done by {0}. Ignoring the message.".format(
                    self.pagure_user
                )
            )
            return

        _log.info("Setting up connection to Pagure")
        self.pagure_io = pagure.set_pagure(config)

        _log.info("Setting up connection to Dist-Git")
        dist_git_config = {
            "pagure_url": config.get("dist_git_url"),
            "pagure_api_key": config.get("dist_git_token"),
        }
        self.dist_git = pagure.set_pagure(dist_git_config)

        _log.info("Setting up connection to FAS")
        fedora_account.set_fasjson(config)

        _log.info("Setting up connection to Bugzilla")
        bugzilla_system.set_bz(config)

        try:
            if message.topic.endswith("pagure.issue.comment.added"):
                self.process_comment(issue)
            else:
                self.process_ticket(issue)
        except BaseException:
            self.pagure_io.add_comment_to_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                comment=(
                    "Error happened during processing:\n" "```\n" "{0}\n" "```\n"
                ).format(traceback.format_exc()),
            )

    def process_comment(self, issue: dict):
        """
        Process comment on the issue.

        Params:
          issue: A dictionary containing the issue.
        """
        _log.info("Processing comment on pagure ticket '{0}'".format(issue["full_url"]))

        last_comment = issue["comments"][-1]

        # Comment is notification
        if last_comment["notification"]:
            _log.info("Comment is notification. Ignoring.")
            return
        # Check if the comment is addressed to bot user
        if "@" + self.pagure_user not in last_comment["comment"]:
            _log.info("Comment is not for '{0}'. Ignoring.".format(self.pagure_user))
            return

        # Extract command from comment @releng-bot command
        command = last_comment["comment"].split()[1]

        if command == "help":
            self.pagure_io.add_comment_to_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                comment=(
                    "{0} recognizes following commands on {1}.\n"
                    "help - prints this help message\n"
                    "retry - retries the failed request processing\n"
                    "valid - can be used by authorized user to mark the request as valid\n"
                ).format(self.pagure_user, PROJECT_NAMESPACE),
            )
        elif command == "retry":
            self.process_ticket(issue)
        elif command == "valid":
            # Get full list of contributors
            contributors = self.pagure_io.get_project_contributors(
                PROJECT_NAMESPACE.split("/")[0], PROJECT_NAMESPACE.split("/")[1]
            )
            # Get the list of maintainers of the scm_requests repository
            maintainers = (
                set(contributors["users"]["admin"])
                | set(contributors["users"]["commit"])
                | set(contributors["users"]["ticket"])
            )
            if last_comment["user"]["name"] in maintainers:
                _log.info(
                    "Ticket is validated by {}, continue processing".format(
                        last_comment["user"]["name"]
                    )
                )
                issue_body_json = json.loads(issue["content"].strip("`").strip("\n"))
                if issue_body_json.get("action") == "new_repo":
                    self._create_new_repo(issue, issue_body_json)
                else:
                    self.pagure_io.add_comment_to_issue(
                        issue["id"],
                        namespace=PROJECT_NAMESPACE,
                        comment=(
                            "Validation can be done for 'new_repo' requests only."
                        ),
                    )
            else:
                self.pagure_io.add_comment_to_issue(
                    issue["id"],
                    namespace=PROJECT_NAMESPACE,
                    comment=(
                        "User {0} is not a maintainer of {1}."
                        "Can't validate this request."
                    ).format(last_comment["user"]["name"], PROJECT_NAMESPACE),
                )

    def process_ticket(self, issue: dict):
        """
        Process a single ticket

        Params:
          issue: A dictionary containing the issue
        """
        _log.info("Handling pagure ticket '{0}'".format(issue["full_url"]))
        try:
            # If a ValueError is raised, that means it isn't valid JSON
            issue_body = json.loads(issue["content"].strip("`").strip("\n"))
        except ValueError:
            _log.info("Invalid JSON in ticket. Closing '{0}'".format(issue["full_url"]))
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message="Invalid JSON provided",
                reason="Invalid",
            )
            return

        if "sls" in issue_body:
            try:
                # If a ValueError is raised, that means they aren't valid SLAs
                _log.info("- Verifying service-levels from the ticket.")
                self.verify_slas(issue_body.get("branch", None), issue_body["sls"])
            except ValidationError as error:
                _log.info(
                    "Couldn't verify SLAs. Closing '{0}'".format(issue["full_url"])
                )
                self.pagure_io.close_issue(
                    issue["id"],
                    namespace=PROJECT_NAMESPACE,
                    message=str(error),
                    reason="Invalid",
                )
                return
        else:
            branch = issue_body.get("branch", "").strip()
            if issue_body.get("action") in ["new_repo", "new_branch"] and branch:
                try:
                    # If the SLAs aren't defined and the branch is a standard
                    # branch, an SLA will be assigned. Otherwise, close this ticket
                    # as bad for missing the SLA.
                    issue_body["sls"] = self.branch_slas[branch]
                except KeyError:
                    _log.info(
                        "Couldn't find standard SLAs. Closing '{0}'".format(
                            issue["full_url"]
                        )
                    )
                    message = "Couldn't find standard SLA for branch '{0}'".format(
                        branch
                    )
                    self.pagure_io.close_issue(
                        issue["id"],
                        namespace=PROJECT_NAMESPACE,
                        message=message,
                        reason="Invalid",
                    )
                    return

        if "branch" in issue_body and "namespace" in issue_body:
            branch = issue_body.get("branch", "")
            ns = issue_body.get("namespace", "")
            if ns in ["flatpaks"]:
                # Validate that branch meets naming standards
                if not re.match(STREAM_NAME_REGEX, branch):
                    error_msg = (
                        "Only characters, numbers, periods, dashes, underscores, "
                        "and pluses are allowed in {} branch names".format(
                            "flatpak" if ns == "flatpaks" else "module"
                        )
                    )
                    self.pagure_io.close_issue(
                        issue["id"],
                        namespace=PROJECT_NAMESPACE,
                        message=error_msg,
                        reason="Invalid",
                    )
                    return

        if (
            "monitor" in issue_body
            and issue_body.get("monitor", "").strip() not in self.monitoring_choices
        ):
            error_msg = 'The monitor choice of "{0}" is invalid'.format(
                issue_body.get("monitor", "")
            )
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message=error_msg,
                reason="Invalid",
            )
            return

        if issue_body.get("action") == "new_repo":
            self.process_new_repo(
                issue, issue_body, initial_commit=issue_body.get("initial_commit", True)
            )
        elif issue_body.get("action") == "new_branch":
            self.create_new_branch(issue, issue_body)
        else:
            _log.info(
                "Unrecognized action '{0}'. Skipping.".format(issue_body.get("action"))
            )
            return

    def verify_slas(self, branch: str, sla_dict: dict):
        """
        Verifies that SLAs are properly formatted

        Params:
          branch: a string of the branch name to verify or None
          sla_dict: a dictionary with the SLAs of the request

        Raises:
          `toddlers.exceptions.ValidationError`: When the sla
            couldn't be validated
        """
        if not isinstance(sla_dict, dict):
            raise ValueError("The object provided is not a dict")

        if branch is not None and branch in self.branch_slas:
            standard_branch_sla_dict = self.branch_slas.get(branch)
            if standard_branch_sla_dict == sla_dict:
                return
            else:
                raise ValidationError(
                    'The SLAs for the branch "{0}" are incorrect'.format(branch)
                )

        for sla, eol in sla_dict.items():
            if not isinstance(eol, str):
                raise ValidationError(
                    'The SL\'s EOL is not a string. It was type "{0}".'.format(
                        type(eol).__name__
                    )
                )
            try:
                eol_date = arrow.get(eol, "YYYY-MM-DD")
            except ValueError:
                raise ValidationError(
                    'The EOL date "{0}" is in an invalid format'.format(eol)
                )
            today = arrow.utcnow()
            if eol_date < today:
                raise ValidationError('The SL "{0}" is already expired'.format(eol))
            elif eol_date.month not in [6, 12] or eol_date.day != 1:
                raise ValidationError(
                    'The SL "{0}" must expire on June 1st or December 1st'.format(eol)
                )

    def process_new_repo(
        self, issue: dict, issue_body_json: dict, initial_commit: bool = True
    ):
        """
        Process request for new repo.

        Params:
          issue: the dictionary representing the issue
          issue_body_json: a partially validated dictionary of the JSON in the
            issue body
          initial_commit: indicate whether to create an initial commit.
        """
        _log.info("New repo was requested. Validating request.")
        required_keys = [
            "repo",
            "branch",
            "bug_id",
            "action",
            "namespace",
            "sls",
            "monitor",
        ]
        for key in required_keys:
            if key not in issue_body_json.keys():
                self.pagure_io.close_issue(
                    issue["id"],
                    namespace=PROJECT_NAMESPACE,
                    message="Invalid body, missing required field: {}".format(key),
                    reason="Invalid",
                )
                return

        # Validate the request first
        if self._validate_new_repo_request(issue, issue_body_json):
            _log.info("Ticket passed all validations. Creating repository.")
            self._create_new_repo(issue, issue_body_json, initial_commit)

    def _validate_new_repo_request(self, issue: dict, issue_body_json: dict) -> bool:
        """
        Validate request for new repo.

        Params:
          issue: the dictionary representing the issue
          issue_body_json: a partially validated dictionary of the JSON in the
            issue body

        Returns:
          True if the request is valid, false otherwise.
        """
        requester = issue["user"]["name"]
        namespace = issue_body_json.get("namespace", "").strip()
        repo = issue_body_json.get("repo", "").strip()
        bug_id = str(issue_body_json.get("bug_id", "")).strip()
        branch_name = issue_body_json.get("branch", "").strip()
        exception = issue_body_json.get("exception", False)

        # Validate the project namespace
        if namespace not in ["rpms", "container", "flatpaks", "tests"]:
            message = (
                "The requested namespace '{0}' is not recognized. "
                "Currently supported namespaces are: "
                "rpms, container, flatpaks, tests.".format(namespace)
            )
            _log.warning(message)
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message=message,
                reason="Invalid",
            )
            return False

        if not re.match(PROJECT_NAME_REGEX, repo) or len(repo) > 64:
            error = (
                "The repository name is invalid. It must be at least two "
                "characters long with only letters, numbers, hyphens, "
                "underscores, plus signs, and/or periods. Please note that "
                "the project cannot start with a period or a plus sign. "
                "Repository name can't be longer than 64 characters."
            )
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message=error,
                reason="Invalid",
            )
            return False

        # This should never trigger because if the user requested an EPEL branch
        # here, it'd have to be accompanied by a Bugzilla bug for the "Fedora EPEL"
        # product. So this will only trigger if the reviewer made a mistake.
        if re.match(EPEL_REGEX, branch_name) and not self.valid_epel_package(
            repo, branch_name
        ):
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message=INVALID_EPEL_ERROR,
                reason="Invalid",
            )
            return False

        _log.info(
            "- Checking if user {0} has an account in dist-git.".format(requester)
        )
        if not self.dist_git.user_exists(requester):
            sync_comment = (
                "@{0} needs to login to {1} to sync accounts "
                "before we can proceed.".format(requester, self.dist_git._pagure_url)
            )
            self.pagure_io.add_comment_to_issue(
                issue["id"], namespace=PROJECT_NAMESPACE, comment=sync_comment
            )
            return False

        # Close the ticket if the requested default branch is master
        if branch_name == "master":
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message="Branch `master` cannot be created, please request the right branch.",
                reason="Invalid",
            )
            return False

        if exception is True or namespace in (
            "flatpaks",
            "tests",
            "container",
        ):
            _log.info("- This ticket needs to be validated manually")
            contributors = self.pagure_io.get_project_contributors(
                PROJECT_NAMESPACE.split("/")[0], PROJECT_NAMESPACE.split("/")[1]
            )
            # Get the list of maintainers of the scm_requests repository
            maintainers = (
                set(contributors["users"]["admin"])
                | set(contributors["users"]["commit"])
                | set(contributors["users"]["ticket"])
            )
            # Ping the maintainers if toddler didn't comment on the ticket yet
            ping_comment = self.ping_comment.format(
                maintainers=" ".join(["@" + user for user in sorted(maintainers)])
            )
            if len(issue["comments"]) > 0:
                _log.info("- Check for ping comment")
                for comment in issue["comments"]:
                    # Skip the notification comments
                    if not comment["notification"]:
                        # We already pinged the maintainers
                        if comment["comment"].strip() == ping_comment:
                            _log.info("- Ping comment found.")
                            return False
            # We didn't commented on the ticket yet, ping the maintainers
            _log.info("- Notify the responsible users")
            self.pagure_io.add_comment_to_issue(
                issue["id"], namespace=PROJECT_NAMESPACE, comment=ping_comment
            )
            return False
        else:
            if not bug_id:
                self.pagure_io.close_issue(
                    issue["id"],
                    namespace=PROJECT_NAMESPACE,
                    message="An invalid Bugzilla bug was provided",
                    reason="Invalid",
                )
                return False
            try:
                _log.info("- Checking that #{0} is a valid RHBZ.".format(bug_id))
                self.validate_review_bug(
                    bug_id,
                    repo,
                    branch_name,
                    namespace=namespace,
                    pagure_user=requester,
                )
            except ValidationError as error:  # pragma: no cover
                self.pagure_io.close_issue(
                    issue["id"],
                    namespace=PROJECT_NAMESPACE,
                    message=str(error),
                    reason="Invalid",
                )
                return False

        return True

    def _create_new_repo(
        self, issue: dict, issue_body_json: dict, initial_commit: bool = True
    ):
        """
        Create new repository.

        Params:
          issue: the dictionary representing the issue
          issue_body_json: a partially validated dictionary of the JSON in the
            issue body
          initial_commit: indicate whether to create an initial commit.
        """
        requester = issue["user"]["name"]
        namespace = issue_body_json.get("namespace", "").strip()
        repo = issue_body_json.get("repo", "").strip()
        bug_id = str(issue_body_json.get("bug_id", "")).strip()
        branch_name = issue_body_json.get("branch", "").strip()
        description = issue_body_json.get("description", "").strip()
        upstreamurl = issue_body_json.get("upstreamurl", "").strip()

        if namespace in ["rpms", "container"]:
            default_branch = "rawhide"
        elif namespace in ["flatpaks"]:
            default_branch = "stable"
        elif namespace in ["tests"]:
            default_branch = "main"

        # Create the Pagure repo
        dist_git_url = "{0}/{1}/{2}".format(
            self.dist_git._pagure_url.rstrip("/"), namespace, repo
        )

        _log.info(
            "- Checking if {0}/{1} already exists in dist-git.".format(namespace, repo)
        )
        project = self.dist_git.get_project(namespace, repo)
        if not project:
            _log.debug("Creating a repo '{0}' in dist-git".format(repo))
            self.dist_git.new_project(
                namespace,
                repo,
                description,
                upstreamurl,
                default_branch,
                initial_commit=initial_commit,
                alias=True,
            )
        else:
            _log.info("Project exists. Skipping to next step.")

        # If the branch requested isn't default branch, create that branch in git. The
        # default branch is already created at this point.
        if branch_name != default_branch:
            _log.info("- Checking if branch exists...")
            branches = self.dist_git.get_branches(namespace, repo)
            if branch_name in branches:
                _log.info(
                    "Branch {0} already exists. Skipping to next step.".format(
                        branch_name
                    )
                )
            else:
                self.dist_git.new_branch(
                    namespace, repo, branch_name, from_branch=default_branch
                )
                _log.info("Branch {0} created.".format(branch_name))

        _log.info(
            "- Setting monitoring status to {0}...".format(
                issue_body_json.get("monitor", "").strip()
            )
        )
        self.dist_git.set_monitoring_status(
            namespace, repo, issue_body_json.get("monitor", "").strip()
        )
        _log.info(
            "Monitoring status set to {0}.".format(
                issue_body_json.get("monitor", "").strip()
            )
        )

        _log.info("- Checking admin of repository to {0}...".format(requester))
        if project and requester in project["access_users"]["owner"]:
            _log.info(
                "Admin of repository already set to {0}. Skipping.".format(requester)
            )
        else:
            self.dist_git.change_project_main_admin(namespace, repo, requester)
            _log.info("Admin of repository set to {0}.".format(requester))

        # Retire rawhide branch for EPEL only packages
        if repo.endswith("-epel"):
            _log.info(
                "Retiring rawhide branch of epel only repository on {0}.".format(repo)
            )
            # First check if the branch isn't already retired
            if self.dist_git.has_dead_package_on_branch(repo, "rawhide"):
                _log.info("Rawhide branch already retired on {0}.".format(repo))
            else:
                with TemporaryDirectory(dir=self.temp_dir) as tmp_dir:
                    dist_git_url = "{0}/{1}/{2}".format(
                        self.dist_git._pagure_url.rstrip("/"), namespace, repo
                    )
                    git_repo = git.clone_repo(dist_git_url, tmp_dir)
                    git_repo.retire_branch(
                        "EPEL only package", self.pagure_user, self.oidc_distgit_token
                    )
                    _log.info("Rawhide branch retired for {0}.".format(repo))

        if branch_name == default_branch:
            new_repo_comment = "The Pagure repository was created at {0}".format(
                dist_git_url
            )
        else:
            new_repo_comment = (
                "The Pagure repository was created at {0}. "
                'You may commit to the branch "{1}" in about '
                "10 minutes.".format(dist_git_url, branch_name)
            )
        self.pagure_io.close_issue(
            issue["id"],
            namespace=PROJECT_NAMESPACE,
            message=new_repo_comment,
            reason="Processed",
        )
        if bug_id:
            bugzilla_system.change_bug_status(
                bug_id, "RELEASE_PENDING", new_repo_comment
            )

    def create_new_branch(self, issue: dict, issue_body_json: dict) -> None:
        """
        A helper function that process request for new branch in dist git.

        Params:
          issue: the dictionary representing the issue
          issue_body_json: a partially validated dictionary of the JSON in the
            issue body
        """
        required_keys = ["action", "namespace", "branch", "sls", "repo"]
        _log.debug("JSON keys in message body {}".format(issue_body_json.keys()))
        for key in required_keys:
            if key not in issue_body_json.keys():
                self.pagure_io.close_issue(
                    issue["id"],
                    namespace=PROJECT_NAMESPACE,
                    message="Invalid body, missing required field: %s" % key,
                    reason="Invalid",
                )
                return

        namespace = issue_body_json.get("namespace", "").strip()
        repo = issue_body_json.get("repo", "").strip()
        bug_id = str(issue_body_json.get("bug_id", "")).strip()
        create_git_branch = issue_body_json.get("create_git_branch", True)

        contributors = self.dist_git.get_project_contributors(namespace, repo)
        if not contributors:
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message="The dist git repository doesn't exist",
                reason="Invalid",
            )
            return

        branch_name = issue_body_json.get("branch", "").strip()
        # Check if the branch doesn't exist in dist git
        branches = self.dist_git.get_branches(namespace, repo)
        if branch_name in branches:
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message="Branch '{}' already exists in dist-git".format(branch_name),
                reason="Invalid",
            )
            return

        if re.match(EPEL_REGEX, branch_name) and not self.valid_epel_package(
            repo, branch_name
        ):
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message=INVALID_EPEL_ERROR,
                reason="Invalid",
            )
            return
        issue_owner = issue["user"]["name"]

        # Check if the branch requestor is one of the maintainers or part of the groups
        _log.info(
            "- Checking if {0} is one of the maintainers of the package".format(
                issue_owner
            )
        )

        # Get the list of maintainers of the package
        maintainers = (
            set(contributors["users"]["admin"])
            | set(contributors["users"]["commit"])
            | set(
                u["user"]
                for u in contributors["users"]["collaborators"]
                if fnmatch.fnmatch(branch_name, u["branches"])
            )
        )
        _log.debug("- Users maintaining the package {0}".format(maintainers))

        # Get the list of FAS groups who can maintain the package
        access_groups = (
            set(contributors["groups"]["admin"])
            | set(contributors["groups"]["commit"])
            | set(
                g["user"]
                for g in contributors["groups"]["collaborators"]
                if fnmatch.fnmatch(branch_name, g["branches"])
            )
        )
        _log.debug("- Groups maintaining the package {0}".format(access_groups))
        group_member = False
        fas_user = fedora_account.get_user_by_username(issue_owner)
        _log.debug("- User requesting new branch {0}".format(fas_user))
        if fas_user:
            for access_group in access_groups:
                # Check if the requestor is part of any of the FAS groups
                # who can maintain the package
                if fedora_account.user_member_of(fas_user, access_group):
                    group_member = True
                    break
        if issue_owner not in maintainers and not group_member:
            ticket_text = (
                "{0} is not a maintainer of the {1} package.\n"
                "Current maintainers of the package.\n"
                "Users:\n"
                "{2}\n"
                "Groups:\n"
                "{3}"
            ).format(
                issue_owner,
                repo,
                "\n".join(["- " + user for user in maintainers]),
                "\n".join(["- " + group for group in access_groups]),
            )
            self.pagure_io.close_issue(
                issue["id"],
                namespace=PROJECT_NAMESPACE,
                message=ticket_text,
                reason="Invalid",
            )
            return

        _log.info("Ticket passed all validations. Creating branch in repository.")

        if create_git_branch:
            with TemporaryDirectory(dir=self.temp_dir) as tmp_dir:
                dist_git_url = "{0}/{1}/{2}".format(
                    self.dist_git._pagure_url.rstrip("/"), namespace, repo
                )
                git_repo = git.clone_repo(dist_git_url, tmp_dir)
                default_branch = self.dist_git.get_default_branch(namespace, repo)
                if not default_branch:
                    # This is a new repo without any branch
                    if not branches:
                        ticket_text = (
                            "The branch couldn't be created in dist git. You can create it by "
                            "yourself as follows:\n"
                            "Check in the project's settings if you have activated "
                            "the git hook preventing "
                            "new git branches from being created and if you did, de-activate it.\n"
                            "Then simply run in cloned repository: "
                            "``git checkout -b <branch_name> && git push -u "
                            "origin <branch_name>``.\n"
                            "``<branch_name>`` is the name of the branch you requested. \n"
                            "You only need to do this once and you can then use "
                            "fedpkg as you normally do."
                        )
                        self.pagure_io.close_issue(
                            issue["id"],
                            namespace=PROJECT_NAMESPACE,
                            message=ticket_text,
                            reason="Invalid",
                        )
                        return
                    else:
                        self.pagure_io.close_issue(
                            issue["id"],
                            namespace=PROJECT_NAMESPACE,
                            message="Default branch is not set for repository '{}/{}'".format(
                                namespace, repo
                            ),
                            reason="Invalid",
                        )
                        return
                commit = git_repo.first_commit(default_branch)
                if commit:
                    # Check if the plugin preventing branch creation is enabled
                    plugins = self.dist_git.get_plugins(namespace, repo)
                    plugin_enabled = False
                    # Plugin is enabled, disabling it
                    if PLUGIN_NAME in plugins:
                        plugin_enabled = True
                        self.dist_git.disable_plugin(namespace, repo, PLUGIN_NAME)
                    try:
                        self.dist_git.new_branch(
                            namespace, repo, branch_name, from_commit=commit
                        )
                    finally:
                        # Re-enable the plugin
                        if plugin_enabled:
                            self.dist_git.enable_plugin(namespace, repo, PLUGIN_NAME)
            new_branch_comment = (
                "The branch was created in git. It "
                "may take up to 10 minutes before you have "
                "write access on the branch."
            )
        else:
            new_branch_comment = (
                "Pagure is still processing "
                "the request, but in about 10 minutes, you may create the "
                "branch in Pagure using git."
            )
        self.pagure_io.close_issue(
            issue["id"],
            namespace=PROJECT_NAMESPACE,
            message=new_branch_comment,
            reason="Processed",
        )
        if bug_id:
            bugzilla_system.comment_on_bug(bug_id, new_branch_comment)

    def validate_review_bug(
        self,
        bug_id: str,
        pkg: str,
        branch: str,
        namespace: str = "rpms",
        pagure_user: Optional[str] = None,
    ):
        """
        Checks the validity of a Bugzilla bug representing a Fedora package
        review. This function was inspired by:
        https://github.com/fedora-infra/pkgdb2/blob/main/pkgdb2/api/extras.py
        https://pagure.io/pkgdb-cli/blob/main/f/pkgdb2client/utils.py

        Params:
          bug_id: ID of the bugzillla bug
          pkg: string of the package name
          branch: string of the branch name
          namespace: a string of the requested repo's namespace
          pagure_user: a string of the requesting user's Pagure username
        """
        try:
            bug = bugzilla_system.get_bug(bug_id)
        except Exception as error:
            raise ValidationError(
                "The Bugzilla bug could not be verified. The following "
                "error was encountered: {0}".format(str(error))
            )
        if not bug:
            raise ValidationError("The Bugzilla bug doesn't exist.")
        # Check that the bug is valid
        bz_proper_component = self.pagure_namespace_to_component.get(namespace)
        bz_proper_products = self.pagure_namespace_to_product.get(namespace) or []
        if namespace == "rpms" and (branch != "rawhide" and branch != "main"):
            if re.match(EPEL_REGEX, branch) and bug.product != "Fedora EPEL":
                raise ValidationError(
                    'The Bugzilla bug is for "{0}" but the '
                    "requested branch is an EPEL branch".format(bug.product)
                )
            elif not re.match(EPEL_REGEX, branch) and bug.product == "Fedora EPEL":
                raise ValidationError(
                    'The Bugzilla bug is for "Fedora EPEL" but the '
                    'requested branch is "{0}"'.format(branch)
                )
        if bz_proper_component is None or bug.component != bz_proper_component:
            raise ValidationError("The Bugzilla bug provided is not the proper type")
        elif bug.product not in bz_proper_products:
            raise ValidationError(
                'The Bugzilla bug provided is not for "{0}"'.format(
                    '" or "'.join(bz_proper_products)
                )
            )
        elif bug.assigned_to in ["", None, "nobody@fedoraproject.org"]:
            raise ValidationError("The Bugzilla bug provided is not assigned to anyone")
        fas_submitter = fedora_account.get_user_by_email(bug.creator)
        if not fas_submitter:
            raise ValidationError(
                "The Bugzilla review bug creator could not be found in "
                "FAS. Make sure your FAS email address is the same as in "
                "Bugzilla."
            )
        if pagure_user:
            if fas_submitter["username"] != pagure_user:
                raise ValidationError(
                    "The Bugzilla review bug creator "
                    "didn't match the requester in Pagure."
                )
        # Check if the review was approved and by whom
        flag_set = False
        for flag in bug.flags:
            if flag.get("name") == "fedora-review":
                if flag.get("status") == "+":
                    flag_set = True
                fas_reviewer = fedora_account.get_user_by_email(bug.assigned_to)
                if not fas_reviewer:
                    raise ValidationError(
                        'The email address "{0}" of the Bugzilla reviewer '
                        "is not tied to a user in FAS or FAS check failed. "
                        "Group membership can't be validated.".format(bug.assigned_to)
                    )
                if not fedora_account.user_member_of(fas_reviewer, "packager"):
                    raise ValidationError(
                        "The Bugzilla bug's review "
                        'is approved by a user "{0}" that is '
                        "not a packager or FAS check failed".format(bug.assigned_to)
                    )
                if not fedora_account.user_member_of(fas_submitter, "packager"):
                    raise ValidationError(
                        'The Bugzilla reporter "{0}"'
                        "is not a packager".format(bug.creator)
                    )

                if flag["setter"] == bug.creator:
                    message = (
                        "The Bugzilla bug's review is approved "
                        "by the person creating the bug. This is "
                        "not allowed."
                    )
                    raise ValidationError(message)

                assigned_to_emails = [bug.assigned_to]

                if "emails" in fas_reviewer:
                    assigned_to_emails.append(fas_reviewer["emails"])

                if flag["setter"] not in assigned_to_emails:
                    raise ValidationError(
                        "The review is not approved by "
                        "the assignee of the Bugzilla "
                        "bug"
                    )

                update_dt = flag.get("modification_date")
                if update_dt:
                    dt = arrow.get(update_dt.value, "YYYYMMDDTHH:mm:ss")
                    delta = arrow.utcnow().date() - dt.date()
                    if delta.days > 60:
                        raise ValidationError(
                            "The Bugzilla bug's review " "was approved over 60 days ago"
                        )
                break
        if not flag_set:
            raise ValidationError("The Bugzilla bug is not approved yet")
        # Check the format of the Bugzilla bug title
        tmp_summary = bug.summary.partition(":")[2]
        if not tmp_summary:
            raise ValidationError(
                'Invalid title for this Bugzilla bug (no ":" present)'
            )
        if " - " not in tmp_summary:
            raise ValidationError(
                'Invalid title for this Bugzilla bug (no "-" present)'
            )
        pkg_in_bug = tmp_summary.split(" - ", 1)[0].strip()
        if pkg != pkg_in_bug:
            message = (
                'The package in the Bugzilla bug "{0}" doesn\'t match '
                'the one provided "{1}"'.format(pkg_in_bug, pkg)
            )
            raise ValidationError(message)

    def valid_epel_package(self, name: str, branch: str) -> bool:
        """
        Determines if the package is allowed to have an EPEL branch.

        Params:
          name: a string of the package name
          branch: a string of the EPEL branch name (e.g. epel9)

        Returns:
          If package is valid EPEL package.

        Raises:
          `ValidationError`: When we can't retrieve list of official EL packages.
        """
        # Extract the major version
        major_version = int(re.findall(r"^epel(\d+)(?:\.\d+)?(?:-next)?$", branch)[0])

        # Starting with epel9 and epel9-next, check against CentOS compose metadata.
        if major_version >= 9:
            # https://bugzilla.redhat.com/show_bug.cgi?id=2005139
            if major_version == 9:
                url = "https://composes.stream.centos.org/production/latest-CentOS-Stream/compose/metadata/rpms.json"  # NOQA E501
            else:
                url = f"https://composes.stream.centos.org/stream-{major_version}/production/latest-CentOS-Stream/compose/metadata/rpms.json"  # NOQA E501
            response = self.requests_session.get(url)

            if response.status_code != 200:
                raise ValidationError(
                    "Couldn't retrieve the list of official EL packages"
                )
            rv_json = response.json()
            pkg_names = {
                # convert name-epoch:version-release.arch key to just the name
                nevra.rsplit(".", maxsplit=1)[0].rsplit("-", maxsplit=2)[0]
                # only packages from these repos are forbidden in epel
                for repo in ["BaseOS", "AppStream", "CRB"]
                for arch in rv_json["payload"]["rpms"][repo].keys()
                for nevra in rv_json["payload"]["rpms"][repo][arch].keys()
            }
            if name in pkg_names:
                return False
        else:
            url = f"https://infrastructure.fedoraproject.org/repo/json/pkg_el{major_version}.json"
            response = self.requests_session.get(url)

            if response.status_code != 200:
                raise ValidationError(
                    "Couldn't retrieve the list of official EL packages"
                )
            rv_json = response.json()

            if name in rv_json["packages"].keys():
                return False
        return True


def _get_arguments(args):
    """Load and parse the CLI arguments.

    :arg args: Script arguments

    :returns: Parsed arguments
    """
    parser = argparse.ArgumentParser(
        description="Processor for SCM requests, handling tickets from '{}'".format(
            PROJECT_NAMESPACE
        )
    )

    parser.add_argument("ticket", type=int, help="Number of ticket to process")

    parser.add_argument(
        "--config",
        help="Configuration file",
    )

    parser.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.INFO,
        help="Enable debugging output",
    )

    return parser.parse_args(args)


def _setup_logging(log_level: int) -> None:
    """Setup the logging level.

    :arg log_level: Log level to set
    """
    handlers = []

    _log.setLevel(log_level)
    # We want all messages logged at level INFO or lower to be printed to stdout
    info_handler = logging.StreamHandler(stream=sys.stdout)
    handlers.append(info_handler)

    if log_level == logging.INFO:
        # In normal operation, don't decorate messages
        for handler in handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    logging.basicConfig(level=log_level, handlers=handlers)


def main(args):
    """Main function."""
    args = _get_arguments(args)
    _setup_logging(log_level=args.log_level)

    with open(args.config, "rb") as conf_fh:
        config = tomllib.load(conf_fh)
    parsed_config = config.get("consumer_config", {}).get("default", {})
    parsed_config.update(
        config.get("consumer_config", {}).get("scm_request_processor", {})
    )

    ticket = args.ticket

    pagure_io = pagure.set_pagure(parsed_config)
    issue = pagure_io.get_issue(ticket, PROJECT_NAMESPACE)

    # Convert issue to message
    body = {"project": {"fullname": PROJECT_NAMESPACE}, "issue": issue}
    message = IssueNewV1(body=body)
    _log.debug("Message prepared: {}".format(message.body))

    SCMRequestProcessor().process(
        config=parsed_config,
        message=message,
    )


if __name__ == "__main__":  # pragma: no cover
    try:
        main(sys.argv[1:])
    except KeyboardInterrupt:
        pass
