from typing import Any, List, Mapping, Optional

from fasjson_client import Client
from fasjson_client.errors import ClientError


# Have a global connection to FAS open.
_FASJSON = None


def set_fasjson(conf: Mapping[str, str]) -> Client:
    """Set the connection to the Fedora Account System through Fasjson."""
    global _FASJSON

    fas_url = conf.get("fas_url")
    if not fas_url:
        raise ValueError("No fas_url found in the configuration file")

    # create a fasjson client
    try:
        _FASJSON = Client(url=conf.get("fas_url"))
    except ClientError as e:
        raise ClientError(message="{0!s}".format(e), code=e.code)
    return _FASJSON


def get_fasjson() -> Client:
    """Retrieve an instantiated FASJSON Client."""
    global _FASJSON
    if _FASJSON is None:
        raise ValueError("No FASJSON client instantiated, call set_fasjson first")

    return _FASJSON


def _get_fas_grp_member(group: str = "packager") -> List[Any]:
    """Retrieve from FAS the list of users in the packager group."""
    fasjson = get_fasjson()
    try:
        return fasjson.list_group_members(groupname=group).result or []
    except ClientError:
        return []


def get_group_member(group_name: str) -> set:
    """Return a list containing the name the members of the given group."""
    output = set()
    for user in _get_fas_grp_member(group_name):
        output.add(user.get("username"))
    return output


def get_bz_email_user(
    username: str, email_overrides: Mapping[str, str]
) -> Optional[str]:
    """Retrieve the bugzilla email associated to the provided username."""
    fasjson = get_fasjson()
    try:
        user_info = fasjson.get_user(username=username).result or {}
        bz_email = user_info.get("rhbzemail", None)
        if not bz_email:
            emails = user_info.get("emails", [])
            if emails:
                bz_email = emails[0]
    except ClientError:
        return None

    if bz_email is None or not isinstance(bz_email, str):
        return None

    bz_email = bz_email.lower().strip()
    bz_email = email_overrides.get(bz_email, bz_email)
    return str(bz_email.strip())


def get_bz_email_group(
    groupname: str, email_overrides: Mapping[str, str]
) -> Optional[str]:
    """Retrieve the bugzilla email associated to the provided group name."""
    fasjson = get_fasjson()
    try:
        group = fasjson.get_group(groupname=groupname).result or {}
    except ClientError:
        return None

    bz_email = group.get("mailing_list")
    if bz_email is None or not isinstance(bz_email, str):
        return None

    bz_email = bz_email.lower().strip()
    ov_email = email_overrides.get(bz_email, bz_email)
    return ov_email.strip()


def get_user_by_email(email: str) -> Optional[dict]:
    """Returns the user found in FAS for that email address."""
    fasjson = get_fasjson()
    user = None
    try:
        result = fasjson.search(email=email).result or []
        if not result:
            # Try the bugzilla email as well
            result = fasjson.search(rhbzemail=email).result or []
        if result:
            user = result[0]
    except ClientError:
        pass

    return user


def user_member_of(user: dict, group: str) -> bool:
    """
    Check if the user is member of group.

    Params:
      user: Dictionary containing user data
      group: Group to check for membership

    Returns:
      Result of group membership. False if the check fails.
    """
    fasjson = get_fasjson()
    is_member = False
    try:
        is_member = fasjson.check_membership(
            groupname=group, username=user["username"]
        ).result
    except ClientError:
        pass

    return is_member


def get_user_by_username(username: str) -> Optional[dict]:
    """
    Returns the user found in FAS for provided username.

    Params:
      username: Username to look for

    Returns:
      Dictionary representing user.
    """
    fasjson = get_fasjson()
    user = None
    try:
        user = fasjson.get_user(username=username).result
    except ClientError:
        pass

    return user
