"""
This module provides the functionality to download the latest primary.xml
database from koji on the rawhide repo.
Decompress that xml file (which are downloaded compressed).
Read its content and build a dictionary with the package names as keys
and their summaries as values.

This code can then be used to create an in-memory cache of this information
which can then later be re-used in other places.
This prevents relying on remote services such as mdapi (of which a lot of
code here is coming from) when needing to access the summary of a lot of
packages.
"""

import contextlib
import hashlib
import logging
import os
import time
from typing import Mapping, Optional
import xml

from defusedxml import ElementTree as etree

from .requests import make_session


repomd_xml_namespace = {
    "repo": "http://linux.duke.edu/metadata/repo",
    "rpm": "http://linux.duke.edu/metadata/rpm",
}

log = logging.getLogger(__name__)


class PackageSummaries:
    """
    Class for retrieving package summaries from koji.
    """

    def __init__(self):
        """
        Constructor.
        """
        self.requests_session = make_session()

    def _download_db(self, repomd_url: str, archive: str) -> None:
        """Download db file.

        Params:
            repomd_url (str): Repository metadata URL
            archive (str): Download destination
        """
        log.info("Downloading file: %s to %s", repomd_url, archive)
        response = self.requests_session.get(repomd_url, verify=True)
        with open(archive, "wb") as stream:
            stream.write(response.content)

    def _decompress_db(self, archive: str, location: str) -> None:
        """Decompress the given archive at the specified location.

        Params:
            archive (str): Path to archive to extract
            location (str): Destination path

        Raises:
            NotImplementedError - when archive format is not supported
        """
        log.info("Extracting %s to %s", archive, location)
        if archive.endswith(".xz"):
            import lzma

            with contextlib.closing(lzma.LZMAFile(archive)) as stream_xz:
                data = stream_xz.read()
            with open(location, "wb") as stream:
                stream.write(data)
        elif archive.endswith(".tar.gz"):
            import tarfile

            with tarfile.open(archive, "r:gz") as tar:
                # Let's assume that there is only one file in archive
                tarinfo = tar.next()
                if tarinfo:
                    tar.extract(tarinfo, path=os.path.dirname(location), filter="data")
                    os.rename(
                        os.path.join(os.path.dirname(location), tarinfo.name), location
                    )
        elif archive.endswith(".gz"):
            import gzip

            with open(location, "wb") as out:
                with gzip.open(archive, "rb") as inp:
                    out.write(inp.read())
        elif archive.endswith(".bz2"):
            import bz2

            with open(location, "wb") as out:
                bzar = bz2.BZ2File(archive)
                out.write(bzar.read())
                bzar.close()
        elif archive.endswith(".zst"):
            import zstandard as zstd

            dctx = zstd.ZstdDecompressor()

            with open(location, "wb") as out, open(archive, "rb") as zstd_file:
                dctx.copy_stream(zstd_file, out)

        else:
            raise NotImplementedError(archive)

    def _needs_update(self, local_file: str, remote_sha: str, sha_type: str) -> bool:
        """Compare hash of a local and remote file.
        Return True if our local file needs to be updated.

        Params:
            local_file: Local file
            remote_sha: Hash of the remote file
            sha_type: Type of the SHA hashing algorithm

        Returns:
            True if our local file needs to be updated.
        """

        if not os.path.isfile(local_file):
            # If we have never downloaded this before, then obviously it has
            # "changed"
            return True

        # Old epel5 doesn't even know which sha it is using...
        if sha_type == "sha":
            sha_type = "sha1"

        hashobj = getattr(hashlib, sha_type)()
        with open(local_file, "rb") as f:
            hashobj.update(f.read())

        local_sha = hashobj.hexdigest()
        if local_sha != remote_sha:
            return True

        return False

    def get_primary_xml(self, destfolder: str, url: str) -> Optional[str]:
        """Retrieve the repo metadata at the given url and store them in destination
        folder.

        Params:
            destfolder: Path to destination folder
            url: URL to download the metadata from

        Returns:
            None or filepath to repository.xml
        """
        repomd_url = url + "/repomd.xml"
        response = self.requests_session.get(repomd_url, verify=True)
        if not bool(response):
            log.warning("Failed to get %s %s", repomd_url, response)
            return None

        try:
            root = etree.fromstring(response.text)
        except xml.etree.ElementTree.ParseError:  # type: ignore
            # mypy was complaining that it can't find etree
            log.warning("Failed to parse %s %s", repomd_url, response)
            return None

        data_nodes = list(
            root.findall('repo:data[@type="primary"]', repomd_xml_namespace)
        )
        if not data_nodes:
            log.debug("No primary.xml could be found in %s", url)
            return None
        if len(data_nodes) > 1:
            log.debug("More than one primary.xml could be found in %s", url)
            return None

        primary_node = data_nodes[0]

        location_node = primary_node.find("repo:location", repomd_xml_namespace)
        if location_node is None or "href" not in location_node.attrib:
            log.debug("No valid location found for primary.xml in %s", url)
            return None

        cksuminfo_node = primary_node.find("repo:open-checksum", repomd_xml_namespace)
        if cksuminfo_node is None or "type" not in cksuminfo_node.attrib:
            log.debug("No valid checksum information found for primary.xml in %s", url)
            return None

        filename = location_node.attrib["href"].replace("repodata/", "")
        hash_digest = cksuminfo_node.text
        hash_type = cksuminfo_node.attrib["type"]

        repomd_url = url + "/" + filename

        # First, determine if the file has changed by comparing hash
        db = "distgit-bugzilla-sync-primary.xml"

        # Have we downloaded this before?  Did it change?
        destfile = os.path.join(destfolder, db)
        if not self._needs_update(destfile, hash_digest, hash_type):
            log.debug("No change of %s", repomd_url)
        else:
            # If it has changed, then download it and move it into place.
            archive = os.path.join(destfolder, filename)

            self._download_db(repomd_url, archive)
            self._decompress_db(archive, destfile)

        return destfile

    def get_package_summaries(self, config: Mapping[str, str]) -> Mapping[str, str]:
        """Get package summaries from XML package db file.

        Params:
          config: Configuration dictionary

        Returns:
          None or dictionary containing package summaries with name of the package as key.
        """
        summaries: Mapping[str, str] = {}

        start = time.time()

        primary_xml = self.get_primary_xml(
            config["temp_folder"],
            config["kojipkgs_url"] + "/repos/rawhide/latest/x86_64/repodata",
        )

        if not primary_xml:
            return summaries

        context = etree.iterparse(primary_xml, events=("start", "end"))

        root = None

        # iterate over the rest of the primary.xml tree
        for event, elem in context:
            if root is None:
                root = elem
                continue

            if (
                event == "end"
                and elem.tag == "{http://linux.duke.edu/metadata/common}package"
                and elem.get("type", "rpm") == "rpm"
            ):
                name = elem.findtext("{http://linux.duke.edu/metadata/common}name")
                summary = elem.findtext(
                    "{http://linux.duke.edu/metadata/common}summary"
                )
                if name is not None and summary is not None:
                    summaries[name] = summary
                # remove package child from root element to keep memory consumption low
                root.clear()

        delta = time.time() - start
        log.info("Parsed in %s seconds -- ie: %s minutes", delta, delta / 60)

        return summaries


if __name__ == "__main__":  # pragma: no cover
    logging.basicConfig(level=logging.DEBUG)
    ps = PackageSummaries()
    dummy_config = {
        "kojipkgs_url": "https://kojipkgs.fedoraproject.org",
        "temp_folder": "/var/tmp",
    }
    result = ps.get_package_summaries(dummy_config)
    print(result)
