"""
This module is a wrapper for pagure. It uses the pagure API to communicate with
configured pagure instance.
To work with it you need to set it up first by calling `set_pagure`.

Examples:
  from utils import pagure, requests

  pagure_config = {
    "pagure_url": "https://pagure.io",
    "pagure_api_key": "secret API key"
  }

  pagure_io = pagure.set_pagure(pagure_config)

  pagure_io.close_issue("<repo>", <issue_id>)
"""

import json
import logging
from typing import Any, List, Optional
import urllib

from toddlers.exceptions import PagureError
from toddlers.utils import requests
from toddlers.utils.cache import cache

log = logging.getLogger(__name__)


def set_pagure(conf: dict) -> Any:
    """
    Set the connection to pagure.

    Params:
      conf: This needs to have keys "pagure_url" and "pagure_api_key"
    """
    return Pagure(conf)


ORPHAN_USERNAME = "orphan"


class Pagure:
    """
    Object that is a wrapper above pagure API.
    """

    # URL to pagure
    _pagure_url: str = ""
    # API key for pagure instance
    _pagure_api_key: Optional[str] = None
    # Requests Session object used for communication
    _requests_session: requests.requests.Session

    def __init__(self, config: dict) -> None:
        """
        Initialize pagure instance.

        Params:
          config: This needs to have keys "pagure_url" and "pagure_api_key"

        Raises:
          `ValueError` if configuration value is missing
        """
        self._pagure_url = config.get("pagure_url", "")
        if not self._pagure_url:
            raise ValueError("No pagure_url found in the configuration file")

        self._pagure_api_key = config.get("pagure_api_key")
        if not self._pagure_api_key:
            raise ValueError("No pagure_api_key found in the configuration file")

        # The API calls could sometimes take more than the default 5 seconds
        # let's increase this to 300
        self._requests_session = requests.make_session(timeout=300)

    def get_auth_header(self) -> dict:
        """
        Get the HTTP header with authorization for the desired service.

        Returns:
          A dictionary of the HTTP header
        """
        return {
            "Authorization": "token {0}".format(self._pagure_api_key),
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

    def get_issue(self, issue_id: int, namespace: str) -> dict:
        """
        Get specific issue from project in pagure.

        Params:
          issue_id: Issue to retrieve
          namespace: Project namespace. For example: 'releng/fedora-scm-requests'

        Returns:
          Dictionary containing pagure issue.

        Raises:
          toddlers.utils.exceptions.PagureError when the issue couldn't be retrieved.
        """
        api_url = "{0}/api/0/{1}/issue/{2}".format(
            self._pagure_url, namespace, issue_id
        )
        headers = self.get_auth_header()

        log.debug(
            "Retrieving issue '{0}' from project '{1}'".format(issue_id, namespace)
        )
        response = self._requests_session.get(api_url, headers=headers)

        result = {}

        if response.status_code == 200:
            result = response.json()
            return result

        log.error(
            "Error when retrieving issue '{0}' from project '{1}'. "
            "Got status_code '{2}'.".format(issue_id, namespace, response.status_code)
        )

        response_json = None
        if response.headers.get("content-type") == "application/json":
            response_json = response.json()
            log.error("Received response: {0}".format(response.json()))

        raise PagureError(
            (
                "Couldn't retrieve issue '{0}' from project '{1}'\n\n"
                "Request to '{2}':\n\n"
                "Response:\n"
                "{3}\n\n"
                "Status code: {4}"
            ).format(
                issue_id,
                namespace,
                api_url,
                response_json,
                response.status_code,
            )
        )

    def close_issue(
        self, issue_id: int, namespace: str, message: str, reason: str = "Closed"
    ):
        """
        Close the issue defined by the id with provided message and reason.

        Params:
          issue_id: Identifier of the issue
          namespace: Project namespace. For example: 'releng/fedora-scm-requests'
          message: Message to add when closing the ticket
          reason: Reason to close the ticket with, this needs to be valid reason
            accepted on the project issue tracker. Default: Fixed.

        Raises:
          toddlers.utils.exceptions.PagureError when the issue couldn't be closed.
        """
        issue_url = "{0}/{1}/issue/{2}".format(self._pagure_url, namespace, issue_id)
        api_url = "{0}/api/0/{1}/issue/{2}/status".format(
            self._pagure_url, namespace, issue_id
        )
        status_payload = {"status": reason}
        headers = self.get_auth_header()

        if message:
            self.add_comment_to_issue(issue_id, namespace=namespace, comment=message)

        log.debug(
            "Closing issue '{0}' with reason '{1}' adding message '{2}'".format(
                issue_url, reason, message
            )
        )
        response = self._requests_session.post(
            api_url, data=json.dumps(status_payload), headers=headers
        )

        if response.status_code == 200:
            return

        log.error(
            "Error when closing issue '{0}'. Got status_code '{1}'.".format(
                issue_url, response.status_code
            )
        )

        response_json = None
        if response.headers.get("content-type") == "application/json":
            response_json = response.json()
            log.error("Received response: {0}".format(response.json()))

        raise PagureError(
            (
                "Couldn't close issue '{0}'\n\n"
                "Request to '{1}':\n"
                "{2}\n\n"
                "Response:\n"
                "{3}\n\n"
                "Status code: {4}"
            ).format(
                issue_url,
                api_url,
                status_payload,
                response_json,
                response.status_code,
            )
        )

    def add_comment_to_issue(self, issue_id: int, namespace: str, comment: str) -> None:
        """
        Comment on issue defined by the id.

        Params:
          issue_id: Identifier of the issue
          namespace: Project namespace. For example: 'releng/fedora-scm-requests'
          message: Message to add when closing the ticket

        Raises:
          toddlers.utils.exceptions.PagureError: When the issue couldn't be commented on.
        """
        issue_url = "{0}/{1}/issue/{2}".format(self._pagure_url, namespace, issue_id)
        api_url = "{0}/api/0/{1}/issue/{2}/comment".format(
            self._pagure_url, namespace, issue_id
        )
        comment_payload = {"comment": comment}
        headers = self.get_auth_header()

        log.debug(
            "Commenting on issue '{0}' with message '{1}'".format(issue_url, comment)
        )
        response = self._requests_session.post(
            api_url, data=json.dumps(comment_payload), headers=headers
        )

        if response.status_code == 200:
            return

        log.error(
            "Error when commenting on issue '{0}'. Got status_code '{1}'.".format(
                issue_url, response.status_code
            )
        )

        response_json = None
        if response.headers.get("content-type") == "application/json":
            response_json = response.json()
            log.error("Received response: {0}".format(response.json()))

        raise PagureError(
            (
                "Couldn't comment on issue '{0}'\n\n"
                "Request to '{1}':\n"
                "{2}\n\n"
                "Response:\n"
                "{3}\n\n"
                "Status code: {4}"
            ).format(
                issue_url,
                api_url,
                comment_payload,
                response_json,
                response.status_code,
            )
        )

    def user_exists(self, username: str) -> bool:
        """
        Check if the user exists in Pagure.

        Params:
        username: a string of the user's username

        Returns:
        Result of the check as boolean.
        """
        user_url = "{0}/api/0/users?pattern={1}".format(self._pagure_url, username)

        response = self._requests_session.get(user_url)

        if response.status_code == 200:
            data = response.json()
            if not data["users"]:
                return False
            return username in data["users"]

        log.error(
            "Error when checking for user '{0}'. Got status_code '{1}'.".format(
                username, response.status_code
            )
        )

        response_json = None
        if response.headers.get("content-type") == "application/json":
            response_json = response.json()
            log.error("Received response: {0}".format(response.json()))

        raise PagureError(
            (
                "Couldn't get user '{0}'\n\n"
                "Request to '{1}':\n"
                "Response:\n"
                "{2}\n\n"
                "Status code: {3}"
            ).format(
                username,
                user_url,
                response_json,
                response.status_code,
            )
        )

    def new_project(
        self,
        namespace: str,
        repo: str,
        description: str,
        upstream_url: str,
        default_branch: str,
        initial_commit: bool = False,
        alias: bool = False,
    ) -> None:
        """
        Create mew project in Pagure.

        Params:
          namespace: Project namespace
          repo: Project name
          description: Description of the project
          upstream_url: URL to upstream of the project
          default_branch: Default branch to create
          initial_commit: Indicate whether to create initial_commit or not
          alias: Indicate whether to create branch alias. Creates alias main->rawhide.
            Use only in case of Dist Git!

        Raises:
            `toddlers.utils.exceptions.PagureError``: When the project creation fails.
        """
        pagure_new_project_url = "{0}/api/0/new".format(self._pagure_url)
        headers = self.get_auth_header()
        payload = {
            "namespace": namespace,
            "name": repo,
            "default_branch": default_branch,
            "description": description or "The {0} package\n".format(repo),
            "url": upstream_url or "",
            "wait": True,
        }

        if initial_commit:
            payload["create_readme"] = True

        log.debug("Creating project '{0}/{1}'".format(namespace, repo))
        response = self._requests_session.post(
            pagure_new_project_url, data=json.dumps(payload), headers=headers
        )

        if response.status_code != 200:
            log.error(
                "Error when creating project '{0}/{1}'. Got status_code '{2}'.".format(
                    namespace, repo, response.status_code
                )
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't create project '{0}/{1}'\n\n"
                    "Request to '{2}':\n"
                    "{3}\n\n"
                    "Response:\n"
                    "{4}\n\n"
                    "Status code: {5}"
                ).format(
                    namespace,
                    repo,
                    pagure_new_project_url,
                    payload,
                    response_json,
                    response.status_code,
                )
            )

        # Don't create alias without initial commit
        # Alias needs to be created from commit
        if alias and initial_commit:
            # Only create alias for rpms and container namespace
            if namespace in ["rpms", "container"]:
                pagure_new_git_alias_url = "{0}/api/0/{1}/{2}/git/alias/new".format(
                    self._pagure_url, namespace, repo
                )
                payload = {
                    "alias_from": "main",
                    "alias_to": "rawhide",
                }
                log.debug(
                    "Creating alias for project '{0}/{1}'".format(namespace, repo)
                )
                response = self._requests_session.post(
                    pagure_new_git_alias_url, data=json.dumps(payload), headers=headers
                )

                if response.status_code != 200:
                    log.error(
                        "Error when creating alias for project '{0}/{1}'. "
                        "Got status_code '{2}'.".format(
                            namespace, repo, response.status_code
                        )
                    )

                    response_json = None
                    if response.headers.get("content-type") == "application/json":
                        response_json = response.json()
                        log.error("Received response: {0}".format(response.json()))

                    raise PagureError(
                        (
                            "Couldn't create alias for project '{0}/{1}'\n\n"
                            "Request to '{2}':\n"
                            "{3}\n\n"
                            "Response:\n"
                            "{4}\n\n"
                            "Status code: {5}"
                        ).format(
                            namespace,
                            repo,
                            pagure_new_git_alias_url,
                            payload,
                            response_json,
                            response.status_code,
                        )
                    )

    def new_branch(
        self,
        namespace: str,
        repo: str,
        branch: str,
        from_commit: str = "",
        from_branch: str = "",
    ) -> None:
        """
        Create a new branch in pagure repository.

        Params:
          namespace: Namespace of the project
          repo: Name of the project to create branch in
          branch: Name of the new branch
          from_commit: Commit from which to create the new branch. Only one of
            `from_commit` or `from_branch` must be set.
          from_commit: Branch from which to create the new branch. Only one of
            `from_commit` or `from_branch` must be set.

        Raises:
            `toddlers.utils.exceptions.PagureError``: When the branch creation fails.
        """
        if from_commit and from_branch:
            raise RuntimeError(
                "`from_commit` and `from_branch` were both " "specified. Only use one."
            )
        elif not from_commit and not from_branch:
            raise RuntimeError(
                "You must specify either `from_commit` or " "`from_branch`"
            )

        branch_api_url = "{0}/api/0/{1}/{2}/git/branch".format(
            self._pagure_url, namespace, repo
        )
        headers = self.get_auth_header()
        payload = {"branch": branch}
        if from_commit:
            payload["from_commit"] = from_commit
        else:
            payload["from_branch"] = from_branch
        log.debug("Payload to send: {0}".format(payload))

        log.debug("Creating branch for project '{0}/{1}'".format(namespace, repo))
        response = self._requests_session.post(
            branch_api_url, data=json.dumps(payload), headers=headers
        )

        if response.status_code != 200:
            log.error(
                "Error when creating branch in project '{0}/{1}'. Got status_code '{2}'.".format(
                    namespace, repo, response.status_code
                )
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't create branch in project '{0}/{1}'\n\n"
                    "Request to '{2}':\n"
                    "{3}\n\n"
                    "Response:\n"
                    "{4}\n\n"
                    "Status code: {5}"
                ).format(
                    namespace,
                    repo,
                    branch_api_url,
                    payload,
                    response_json,
                    response.status_code,
                )
            )

    def set_monitoring_status(
        self,
        namespace: str,
        repo: str,
        monitoring_level: str,
    ) -> None:
        """
        Set a monitoring status for pagure repository. This will work only on dist git.

        Params:
          namespace: Namespace of the project
          repo: Name of the project to create branch in
          monitoring_level: Monitoring level to set

        Raises:
            `toddlers.utils.exceptions.PagureError``: When setting the monitoring level fails.
        """
        monitoring_api_url = "{0}/_dg/anitya/{1}/{2}".format(
            self._pagure_url, namespace, repo
        )
        headers = self.get_auth_header()
        payload = {"anitya_status": monitoring_level}

        log.debug(
            "Setting monitoring to '{0}' for project '{1}/{2}'".format(
                monitoring_level, namespace, repo
            )
        )
        response = self._requests_session.post(
            monitoring_api_url, data=json.dumps(payload), headers=headers
        )

        if response.status_code != 200:
            log.error(
                "Error when setting monitoring on project '{0}/{1}'. Got status_code '{2}'.".format(
                    namespace, repo, response.status_code
                )
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't set monitoring on project '{0}/{1}'\n\n"
                    "Request to '{2}':\n"
                    "{3}\n\n"
                    "Response:\n"
                    "{4}\n\n"
                    "Status code: {5}"
                ).format(
                    namespace,
                    repo,
                    monitoring_api_url,
                    payload,
                    response_json,
                    response.status_code,
                )
            )

    def change_project_main_admin(
        self, namespace: str, repo: str, new_main_admin: str
    ) -> None:
        """
        Change the main admin of a project in pagure.

        Params:
          namespace: Namespace of the project
          repo: Name of the project
          new_admin_main: Username of the new main admin

        Raises:
            `toddlers.utils.exceptions.PagureError``: When setting new admin fails.
        """
        admin_api_url = "{0}/api/0/{1}/{2}".format(self._pagure_url, namespace, repo)
        headers = self.get_auth_header()
        payload = {"main_admin": new_main_admin}

        log.debug(
            "Setting new admin to '{0}' for project '{1}/{2}'".format(
                new_main_admin, namespace, repo
            )
        )
        response = self._requests_session.patch(
            admin_api_url, data=json.dumps(payload), headers=headers
        )

        if response.status_code != 200:
            log.error(
                "Error when setting new admin on project '{0}/{1}'. Got status_code '{2}'.".format(
                    namespace, repo, response.status_code
                )
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't set new admin on project '{0}/{1}'\n\n"
                    "Request to '{2}':\n"
                    "{3}\n\n"
                    "Response:\n"
                    "{4}\n\n"
                    "Status code: {5}"
                ).format(
                    namespace,
                    repo,
                    admin_api_url,
                    payload,
                    response_json,
                    response.status_code,
                )
            )

    def get_project_contributors(self, namespace: str, repo: str) -> Any:
        """
        Get contributors of the project.

        Params:
          namespace: Namespace of the project
          repo: Name of the project

        Returns:
          Dictionary containing list of contributors or None if the project doesn't exist.

        Raises:
          `toddlers.utils.exceptions.PagureError``: When getting contributors fails.
        """
        contributors_api_url = "{0}/api/0/{1}/{2}/contributors".format(
            self._pagure_url, namespace, repo
        )
        headers = self.get_auth_header()

        log.debug("Getting contributors for project '{0}/{1}'".format(namespace, repo))
        response = self._requests_session.get(contributors_api_url, headers=headers)

        if response.ok:
            result = response.json()
        elif response.status_code == 404:
            result = None
        else:
            log.error(
                "Error when retrieving contributors project '{0}/{1}'. "
                "Got status_code '{2}'.".format(namespace, repo, response.status_code)
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't get contributors for project '{0}/{1}'\n\n"
                    "Request to '{2}':\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    repo,
                    contributors_api_url,
                    response_json,
                    response.status_code,
                )
            )

        return result

    @cache.cache_on_arguments(expiration_time=360)
    def get_branches(self, namespace: str, repo: str) -> Any:
        """
        Return all branches for the specified repository.

        Params:
          namespace: Namespace of the project
          repo: Name of the project

        Returns:
          List of the branches in repository.

        Raises:
          `toddlers.utils.exceptions.PagureError``: When getting branches fails.
        """
        branches_api_url = "{0}/api/0/{1}/{2}/git/branches".format(
            self._pagure_url, namespace, repo
        )
        headers = self.get_auth_header()

        log.debug(
            "Getting list of branches for project '{0}/{1}'".format(namespace, repo)
        )
        response = self._requests_session.get(branches_api_url, headers=headers)

        if response.status_code != 200:
            log.error(
                "Error when getting list of branches for project '{0}/{1}'. "
                "Got status_code '{2}'.".format(namespace, repo, response.status_code)
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't get list of branches for project '{0}/{1}'\n\n"
                    "Request to '{2}':\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    repo,
                    branches_api_url,
                    response_json,
                    response.status_code,
                )
            )

        result = response.json().get("branches", [])

        return result

    @cache.cache_on_arguments(expiration_time=3600)
    def get_default_branch(self, namespace: str, repo: str) -> Any:
        """
        Return the default branch for the specified repository.

        Params:
          namespace: Namespace of the project
          repo: Name of the project

        Returns:
          Name of the default branch or None.

        Raises:
          `toddlers.utils.exceptions.PagureError``: When getting default branch fails.
        """
        branches_api_url = "{0}/api/0/{1}/{2}/git/branches".format(
            self._pagure_url, namespace, repo
        )
        headers = self.get_auth_header()

        log.debug(
            "Getting default branch for project '{0}/{1}'".format(namespace, repo)
        )
        response = self._requests_session.get(branches_api_url, headers=headers)

        if response.status_code != 200:
            log.error(
                "Error when getting default branch for project '{0}/{1}'. "
                "Got status_code '{2}'.".format(namespace, repo, response.status_code)
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't get default branch for project '{0}/{1}'\n\n"
                    "Request to '{2}':\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    repo,
                    branches_api_url,
                    response_json,
                    response.status_code,
                )
            )

        result = response.json().get("default")

        return result

    def get_project(self, namespace: str, repo: str) -> Any:
        """
        Get project.

        Params:
          namespace: Namespace of the project
          repo: Name of the project

        Returns:
          Dictionary containing project or None if project doesn't exists..

        Raises:
          `toddlers.utils.exceptions.PagureError``: When getting project fails.
        """
        contributors_api_url = "{0}/api/0/{1}/{2}".format(
            self._pagure_url, namespace, repo
        )
        headers = self.get_auth_header()

        log.debug("Getting project '{0}/{1}'".format(namespace, repo))
        response = self._requests_session.get(contributors_api_url, headers=headers)

        if response.ok:
            result = response.json()
        elif response.status_code == 404:
            result = None
        else:
            log.error(
                "Error when retrieving project '{0}/{1}'. "
                "Got status_code '{2}'.".format(namespace, repo, response.status_code)
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: {0}".format(response.json()))

            raise PagureError(
                (
                    "Couldn't get project '{0}/{1}'\n\n"
                    "Request to '{2}':\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    repo,
                    contributors_api_url,
                    response_json,
                    response.status_code,
                )
            )

        return result

    @cache.cache_on_arguments(expiration_time=3600)
    def is_project_orphaned(self, namespace: str, repo: str) -> bool:
        """
        Check if project is orphaned.
        Working only with dist-git

        Params:
            namespace: Namespace of project
            repo: Name of the project

        Returns:
            Result of check as boolean.

        Raises:
            `toddlers.utils.exceptions.PagureError``: When getting project maintainers fails.
        """
        endpoint_url = f"{self._pagure_url}/_dg/bzoverrides/{namespace}/{repo}"
        headers = self.get_auth_header()

        log.debug("Getting project '{0}/{1}' maintainers.")
        response = self._requests_session.get(endpoint_url, headers=headers)

        if response.ok:
            result = response.json()
            if (
                result["epel_assignee"] == ORPHAN_USERNAME
                or result["fedora_assignee"] == ORPHAN_USERNAME
            ):
                return True
            else:
                return False
        else:
            log.error(
                "Error when retrieving project '{0}/{1}' maintainers. "
                "Got status_code '{2}'.".format(namespace, repo, response.status_code)
            )
            raise PagureError(
                "Couldn't get project '{0}/{1}' maintainers.".format(namespace, repo)
            )

    def assign_maintainer_to_project(
        self, namespace: str, repo: str, maintainer_fas: str
    ) -> None:
        """
        Assign maintainer to project.
        Working only with dist-git

        Params:
            namespace: Namespace of project
            repo: Name of the project
            maintainer_fas: FAS of user that will maintain.

        Raises:
            `toddlers.utils.exceptions.PagureError``: When setting project maintainer fails.
        """
        endpoint_url = f"{self._pagure_url}/_dg/bzoverrides/{namespace}/{repo}"
        headers = self.get_auth_header()
        payload = {
            "EPEL Maintainer name": maintainer_fas,
            "Maintainer name": maintainer_fas,
        }

        log.debug(
            "Setting new maintainer to '{0}' for project '{1}/{2}'".format(
                maintainer_fas, namespace, repo
            )
        )
        response = self._requests_session.patch(
            endpoint_url, data=json.dumps(payload), headers=headers
        )

        if response.status_code != 200:
            log.error(
                "Error when setting new maintainer on project '{0}/{1}'. Got status_code '{2}'."
                "".format(namespace, repo, response.status_code)
            )
            raise PagureError(
                "Couldn't set new maintainer on project '{0}/{1}'".format(
                    namespace, repo
                )
            )
        if maintainer_fas == ORPHAN_USERNAME:
            self.is_project_orphaned.invalidate(self, namespace, repo)  # type: ignore[attr-defined]

    def add_member_to_group(self, user: str, group: str) -> None:
        """
        Add member to pagure group.

        Params:
            user: User to add
            group: Group to add user to

        Raises:
            `toddlers.utils.exceptions.PagureError``: When user can't be added to group.
        """
        api_url = "{0}/api/0/group/{1}/add".format(self._pagure_url, group)

        payload = {"user": user}
        headers = self.get_auth_header()

        log.debug("Adding user '%s' to group '%s'", user, group)
        response = self._requests_session.post(
            api_url, data=json.dumps(payload), headers=headers
        )

        if not response.ok:
            log.error(
                "Error when adding user '%s' to group '%s'. " "Got status_code '%s'.",
                user,
                group,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't add user '{0}' to group '{1}'\n\n"
                    "Request to '{2}':\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    user,
                    group,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug("User '%s' added to group '%s'", user, group)

    def remove_member_from_group(self, user: str, group: str) -> None:
        """
        Remove member from pagure group.

        Params:
            user: User to remove
            group: Group to remove user from

        Raises:
            `toddlers.utils.exceptions.PagureError``: When user can't be removed to group.
        """
        api_url = "{0}/api/0/group/{1}/remove".format(self._pagure_url, group)
        payload = {"user": user}
        headers = self.get_auth_header()

        log.debug("Removing user '%s' from group '%s'", user, group)
        response = self._requests_session.post(
            api_url, data=json.dumps(payload), headers=headers
        )

        if not response.ok:
            log.error(
                "Error when removing user '%s' from group '%s'. "
                "Got status_code '%s'.",
                user,
                group,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't remove user '{0}' from group '{1}'\n\n"
                    "Request to '{2}':\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    user,
                    group,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug("User '%s' removed from group '%s'", user, group)

    def get_group_members(self, group: str) -> List[str]:
        """
        Get members of group.

        Params:
          group: Group name

        Returns:
          List of users or empty list if group doesn't exists..

        Raises:
          `toddlers.utils.exceptions.PagureError``: When getting members fail.
        """
        result = []
        api_url = "{0}/api/0/group/{1}".format(self._pagure_url, group)
        headers = self.get_auth_header()

        log.debug("Getting members of group '%s'", group)
        response = self._requests_session.get(api_url, headers=headers)

        if response.ok:
            if "members" in response.json():
                result = response.json()["members"]
            else:
                raise PagureError(
                    "The 'members' parameter is missing in response. "
                    "JSON response: {0}".format(response.json())
                )
        elif response.status_code == 404:
            return result
        else:
            log.error(
                "Error when retrieving members from group '%s'. "
                "Got status_code '%s'.",
                group,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't get members of group '{0}'\n\n"
                    "Request to '{1}:\n\n"
                    "Response:\n"
                    "{2}\n\n"
                    "Status code: {3}"
                ).format(
                    group,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug("Members retrieved for group '%s'", group)

        return result

    @cache.cache_on_arguments(expiration_time=7200)
    def get_retired_packages(self, branch: str) -> list:
        """
        Retrieve list of retired packages in specific dist git branch.

        Working only with dist-git

        Params:
            branch: Dist git branch to get the retired packages for. For example "f40".

        Raises:
            `toddlers.utils.exceptions.PagureError``: When retrieving the retired list fails.
        """
        result = []
        endpoint_url = f"{self._pagure_url}/lookaside/retired_in_{branch}.json"
        headers = self.get_auth_header()

        log.debug("Retrieving retired packages for '%s'", branch)
        response = self._requests_session.get(endpoint_url, headers=headers)

        if response.ok:
            try:
                result = response.json()[branch]
            except KeyError:
                raise PagureError(
                    (
                        "Expected key not found in pagure response. "
                        "Expected data: {{<branch_name>: ['package']}}\n\n"
                        "Request to: '{0}'\n\n"
                        "Response:\n"
                        "{1}\n\n"
                        "Status code: {2}"
                    ).format(
                        endpoint_url,
                        response.json(),
                        response.status_code,
                    )
                )
        # When the branch doesn't exist return empty list
        elif response.status_code == 404:
            result = []
        else:
            log.error(
                "Error when retrieving retired packages for '%s'. "
                "Got status_code '%s'.",
                branch,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't retrieve retired packages for '{0}'\n\n"
                    "Request to: '{1}'\n\n"
                    "Response:\n"
                    "{2}\n\n"
                    "Status code: {3}"
                ).format(
                    branch,
                    endpoint_url,
                    response_json,
                    response.status_code,
                )
            )

        return result

    def is_retired_on_branch(self, package: str, branch: str) -> bool:
        """
        Check if the given package is retired on branch.

        Works only with dist-git

        Params:
          package: Package name
          branch: Branch name
        """
        retired_on_branch = self.get_retired_packages(branch)

        if package in retired_on_branch:
            return True
        else:
            return False

    def has_dead_package_on_branch(self, package: str, branch: str) -> bool:
        """
        Check if the given package has dead.package in distgit on particular
        branch. Might look the same as the function `is_retired_on_branch`.
        The difference is that the json files in lookaside reflects the status
        of the packages with a small delay (up to 1 day).

        Works only with distgit.

        Params:
          package: Package name
          branch: Branch name
        """
        distgit_url = f"{self._pagure_url}/api/0/rpms"
        endpoint_url = f"{distgit_url}/{package}/tree/{branch}/f/dead.package"

        headers = self.get_auth_header()

        log.debug(f"Checking if the dead.package is present on branch {branch}")
        response = self._requests_session.get(endpoint_url, headers=headers)

        if response.ok and response.json()["name"] == "dead.package":
            return True
        elif response.status_code == 404:
            return False
        else:
            log.error(
                "Error when checking for dead.package in distgit for '%s'. "
                "Got status_code '%s'.",
                branch,
                response.status_code,
            )
            return False

    def get_all_groups(self):
        """Returns list of all groups from Pagure or Distgit"""
        url = f"{self._pagure_url}/api/0/groups"
        all_groups = []
        while url:
            log.info(f"Retrieving groups from {url}")
            response = self._requests_session.get(url)
            data = response.json()
            groups = data.get("groups", [])
            url = data.get("pagination", dict()).get("next", "")
            all_groups.extend(groups)
        return all_groups

    def orphan_package(
        self, namespace: str, package: str, reason: str, info: str
    ) -> None:
        """
        Give the specified package to 'orphan' user.

        Working only with dist-git

        Params:
          namespace: Namespace of the package
          package: Package name
          reason: Reason for orphaning package
          info: Information why the package is orphaned

        Raises:
          `toddlers.utils.exceptions.PagureError``: When orphaning the package fails.
        """
        api_url = "{0}/_dg/orphan/{1}/{2}".format(self._pagure_url, namespace, package)
        headers = self.get_auth_header()
        payload = {"orphan_reason": reason, "orphan_reason_info": info}

        log.debug("Orphaning package '%s/%s'", namespace, package)
        response = self._requests_session.post(
            api_url, data=json.dumps(payload), headers=headers
        )

        if not response.ok:
            log.error(
                "Error when orphaning package '%s/%s'. " "Got status_code '%s'.",
                namespace,
                package,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't orphan package '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    package,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug("Package '%s/%s' successfully orphaned.", namespace, package)
        self.is_project_orphaned.invalidate(self, namespace, package)  # type: ignore[attr-defined]

    def set_bugzilla_overrides(
        self,
        namespace: str,
        package: str,
        fedora_assignee: Optional[str] = None,
        epel_assignee: Optional[str] = None,
    ) -> None:
        """
        Set the bugzilla overrides for package to specific value.

        Working only with dist-git

        Params:
          namespace: Namespace of the package
          package: Package name
          fedora_assignee: Package maintainer for Fedora
          epel_assignee: Package maintainer for EPEL

        Raises:
          `toddlers.utils.exceptions.PagureError``: When setting bugzilla overrides fails.
        """
        api_url = "{0}/_dg/bzoverrides/{1}/{2}".format(
            self._pagure_url, namespace, package
        )
        headers = self.get_auth_header()
        payload = {"fedora_assignee": fedora_assignee, "epel_assignee": epel_assignee}

        log.debug(
            "Settings bugzilla maintainers for '%s/%s' to fedora: '%s' and epel: '%s'",
            namespace,
            package,
            fedora_assignee,
            epel_assignee,
        )
        response = self._requests_session.post(
            api_url, data=json.dumps(payload), headers=headers
        )

        if not response.ok:
            log.error(
                "Error when setting bugzilla maintainers for '%s/%s'. "
                "Got status_code '%s'.",
                namespace,
                package,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't set bugzilla maintainers for '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    package,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug(
            "Bugzilla maintainers for '%s/%s' changed to fedora: '%s' and epel: '%s'.",
            namespace,
            package,
            fedora_assignee,
            epel_assignee,
        )
        if fedora_assignee == ORPHAN_USERNAME or epel_assignee == ORPHAN_USERNAME:
            self.is_project_orphaned.invalidate(  # type: ignore[attr-defined]
                self,
                namespace,
                package,
            )

    def modify_acls(
        self,
        namespace: str,
        package: str,
        user_type: str,
        name: str,
    ) -> None:
        """
        Modify ACLs on the project.

        Params:
          namespace: Namespace of the package
          package: Package name
          user_type: User or group
          name: Username or group name

        Raises:
          `toddlers.utils.exceptions.PagureError``: When modifying ACLs fails.
        """
        api_url = "{0}/api/0/{1}/{2}/git/modifyacls".format(
            self._pagure_url, namespace, package
        )
        headers = self.get_auth_header()
        payload = {"user_type": user_type, "name": name}

        log.debug(
            "Modify ACLs on '%s/%s' for '%s' '%s'",
            namespace,
            package,
            user_type,
            name,
        )
        response = self._requests_session.post(
            api_url, data=json.dumps(payload), headers=headers
        )

        if not response.ok:
            log.error(
                "Modify ACLs for '%s/%s' failed. " "Got status_code '%s'.",
                namespace,
                package,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't modify ACLs for '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    package,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug(
            "ACLs of '%s/%s' were changed for '%s' '%s'.",
            namespace,
            package,
            user_type,
            name,
        )

    def update_watcher_status(
        self,
        namespace: str,
        package: str,
        status: str,
        watcher: str,
    ) -> None:
        """
        Update watcher status on specified project.

        Params:
          namespace: Namespace of the package
          package: Package name
          status: Watch status to set for watcher
          watcher: Name of the watcher

        Raises:
          `toddlers.utils.exceptions.PagureError``: When updating watcher status fails.
        """
        api_url = "{0}/api/0/{1}/{2}/watchers/update".format(
            self._pagure_url, namespace, package
        )
        headers = self.get_auth_header()
        payload = {"status": status, "watcher": watcher}

        log.debug(
            "Update watcher status on '%s/%s' for '%s' to '%s'",
            namespace,
            package,
            watcher,
            status,
        )
        response = self._requests_session.post(
            api_url, data=json.dumps(payload), headers=headers
        )

        if not response.ok:
            log.error(
                "Update watcher status '%s/%s' failed. " "Got status_code '%s'.",
                namespace,
                package,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Couldn't update watcher status on '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    package,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug(
            "Watcher status on '%s/%s' changed for '%s' to '%s'.",
            namespace,
            package,
            watcher,
            status,
        )

    def get_plugins(
        self,
        namespace: str,
        project: str,
    ) -> List[str]:
        """
        Get enabled plugins on project.

        Params:
          namespace: Namespace of the project
          project: Project name

        Raises:
          `toddlers.utils.exceptions.PagureError``: When getting plugins fails.
        """
        api_url = "{0}/api/0/{1}/{2}/settings/plugins".format(
            self._pagure_url, namespace, project
        )
        headers = self.get_auth_header()

        log.debug(
            "Get enabled plugins on '%s/%s'",
            namespace,
            project,
        )
        response = self._requests_session.get(api_url, headers=headers)

        if not response.ok:
            log.error(
                "Getting enabled plugins on '%s/%s' failed. " "Got status_code '%s'.",
                namespace,
                project,
                response.status_code,
            )

            response_json = None
            print(response.headers)
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Failed to get enabled plugins on '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    project,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        response_json = response.json()
        plugins = []
        for plugin in response_json["plugins"]:
            plugins.extend(plugin.keys())

        log.debug(
            "Retrieved enabled plugins on '%s/%s': %s", namespace, project, plugins
        )

        return plugins

    def disable_plugin(
        self,
        namespace: str,
        project: str,
        plugin: str,
    ) -> None:
        """
        Disable plugin on project.

        Params:
          namespace: Namespace of the project
          project: Project name
          plugin: Plugin to disable

        Raises:
          `toddlers.utils.exceptions.PagureError``: When disabling plugins fails.
        """
        api_url = "{0}/api/0/{1}/{2}/settings/{3}/remove".format(
            self._pagure_url, namespace, project, urllib.parse.quote(plugin)
        )
        headers = self.get_auth_header()

        log.debug(
            "Disable plugin '%s' on '%s/%s'",
            plugin,
            namespace,
            project,
        )
        response = self._requests_session.post(
            api_url, data=json.dumps({}), headers=headers
        )

        if not response.ok:
            log.error(
                "Disabling plugin on '%s/%s' failed. " "Got status_code '%s'.",
                namespace,
                project,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Failed to disable plugin on '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    project,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug("Disabled plugin '%s' on '%s/%s': %s", plugin, namespace, project)

    def enable_plugin(
        self,
        namespace: str,
        project: str,
        plugin: str,
    ) -> None:
        """
        Enable plugin on project.

        Params:
          namespace: Namespace of the project
          project: Project name
          plugin: Plugin to enable

        Raises:
          `toddlers.utils.exceptions.PagureError``: When disabling plugins fails.
        """
        api_url = "{0}/api/0/{1}/{2}/settings/{3}/install".format(
            self._pagure_url, namespace, project, urllib.parse.quote(plugin)
        )
        headers = self.get_auth_header()

        log.debug(
            "Enable plugin '%s' on '%s/%s'",
            plugin,
            namespace,
            project,
        )
        response = self._requests_session.post(
            api_url, data=json.dumps({}), headers=headers
        )

        if not response.ok:
            log.error(
                "Enabling plugin on '%s/%s' failed. " "Got status_code '%s'.",
                namespace,
                project,
                response.status_code,
            )

            response_json = None
            if response.headers.get("content-type") == "application/json":
                response_json = response.json()
                log.error("Received response: %s", response.json())

            raise PagureError(
                (
                    "Failed to enable plugin on '{0}/{1}'\n\n"
                    "Request to '{2}:\n\n"
                    "Response:\n"
                    "{3}\n\n"
                    "Status code: {4}"
                ).format(
                    namespace,
                    project,
                    api_url,
                    response_json,
                    response.status_code,
                )
            )

        log.debug("Enabled plugin '%s' on '%s/%s': %s", plugin, namespace, project)
