"""Create requests session objects with defaults"""

from typing import Any, Optional, Tuple, Union


import requests
from urllib3.util import retry
from urllib3.util.timeout import Timeout as timeout  # type: ignore

# mypy couldn't find the typing information for default

_Default = timeout.DEFAULT_TIMEOUT
__all__ = ["make_session"]

RETRY_DEFAULTS = {
    "total": 3,
    "backoff_factor": 1,
    "BACKOFF_MAX": 5,
}

TIMEOUT_DEFAULT = 5


class TimeoutHTTPAdapter(requests.adapters.HTTPAdapter):
    """A requests adapter with default timeout values"""

    def __init__(
        self,
        *args: list,
        timeout: Optional[Union[Tuple[float], float, object]] = _Default,
        **kwargs: dict,
    ):
        if timeout is not _Default:
            self.timeout = timeout
        else:
            self.timeout = TIMEOUT_DEFAULT
        # mypy is confused with the **kwargs
        super().__init__(*args, **kwargs)  # type: ignore

    # mypy doesn't handle **kwargs wrap well
    def send(self, request: requests.Request, **kwargs: Any) -> requests.Response:  # type: ignore
        kwargs.setdefault("timeout", _Default)
        if kwargs["timeout"] is _Default:
            kwargs["timeout"] = self.timeout
        # mypy is confused with the **kwargs
        return super().send(request, **kwargs)  # type: ignore


class DefaultTimeoutSession(requests.Session):
    """Customize ``requests.Session`` so it works better with default timeouts"""

    def resolve_redirects(self, *args, **kwargs):
        kwargs.setdefault("timeout", _Default)
        return super().resolve_redirects(*args, **kwargs)

    def request(self, *args, **kwargs):
        kwargs.setdefault("timeout", _Default)
        return super().request(*args, **kwargs)


def make_session(
    timeout: Optional[Union[Tuple[float], float, object]] = _Default,
    retry_conf: Optional[dict] = None,
) -> DefaultTimeoutSession:
    """Make a customized requests session object

    Args:
        timeout: a default timeout value
        retry_conf: retry configuration settings, its keys/values will be
            passed as arguments to the constructor of
            :py:class:`requests.packages.urllib3.util.retry.Retry`, or (in the
            case of ``BACKOFF_MAX``), set as an attribute on the created retry
            configuration object.
    Return:
        a session object
    """
    if not retry_conf:
        retry_conf = {}
    else:
        retry_conf = retry_conf.copy()

    for k, v in RETRY_DEFAULTS.items():
        retry_conf.setdefault(k, v)

    BACKOFF_MAX = retry_conf.pop("BACKOFF_MAX")

    retry.Retry.DEFAULT_BACKOFF_MAX = BACKOFF_MAX
    retry_conf_obj = retry.Retry(**retry_conf)
    # retry_conf_obj.BACKOFF_MAX = BACKOFF_MAX

    http_adapter = TimeoutHTTPAdapter(timeout=timeout, max_retries=retry_conf_obj)  # type: ignore

    session = DefaultTimeoutSession()
    for prefix in ("http://", "https://"):
        session.mount(prefix, http_adapter)

    return session
